import pytest

from allotropy.parser_factory import Vendor
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

output_files = (
    "Appbio_AbsoluteQ_example01.csv",
    "Appbio_AbsoluteQ_example02.csv",
    "Appbio_AbsoluteQ_example03.csv",
    "Appbio_AbsoluteQ_example04.csv",
    "Appbio_AbsoluteQ_example05.csv",
)

VENDOR_TYPE = Vendor.APPBIO_ABSOLUTE_Q


@pytest.mark.parametrize("output_file", output_files)
def test_parse_appbio_absolute_q_to_asm_schema(output_file: str) -> None:
    test_filepath = f"tests/parsers/appbio_absolute_q/testdata/{output_file}"
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, "pcr/BENCHLING/2023/09/dpcr.json")


@pytest.mark.parametrize("output_file", output_files)
def test_parse_appbio_absolute_q_to_asm_contents(output_file: str) -> None:
    test_filepath = f"tests/parsers/appbio_absolute_q/testdata/{output_file}"
    expected_filepath = test_filepath.replace(".csv", ".json")
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)

    validate_contents(allotrope_dict, expected_filepath)



from datetime import tzinfo
from enum import Enum
from typing import Optional, Union

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parsers.agilent_gen5.agilent_gen5_parser import AgilentGen5Parser
from allotropy.parsers.appbio_absolute_q.appbio_absolute_q_parser import (
    AppbioAbsoluteQParser,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_parser import (
    AppBioQuantStudioParser,
)
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_parser import ViCellBluParser
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_parser import ViCellXRParser
from allotropy.parsers.moldev_softmax_pro.softmax_pro_parser import SoftmaxproParser
from allotropy.parsers.novabio_flex2.novabio_flex2_parser import NovaBioFlexParser
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_parser import (
    PerkinElmerEnvisionParser,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_parser import (
    RocheCedexBiohtParser,
)
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from allotropy.parsers.vendor_parser import VendorParser


class Vendor(Enum):
    AGILENT_GEN5 = "AGILENT_GEN5"
    APPBIO_QUANTSTUDIO = "APPBIO_QUANTSTUDIO"
    APPBIO_ABSOLUTE_Q = "APPBIO_ABSOLUTE_Q"
    BECKMAN_VI_CELL_BLU = "BECKMAN_VI_CELL_BLU"
    BECKMAN_VI_CELL_XR = "BECKMAN_VI_CELL_XR"
    MOLDEV_SOFTMAX_PRO = "MOLDEV_SOFTMAX_PRO"
    NOVABIO_FLEX2 = "NOVABIO_FLEX2"
    PERKIN_ELMER_ENVISION = "PERKIN_ELMER_ENVISION"
    ROCHE_CEDEX_BIOHT = "ROCHE_CEDEX_BIOHT"


VendorType = Union[Vendor, str]


_VENDOR_TO_PARSER: dict[Vendor, type[VendorParser]] = {
    Vendor.AGILENT_GEN5: AgilentGen5Parser,
    Vendor.APPBIO_QUANTSTUDIO: AppBioQuantStudioParser,
    Vendor.APPBIO_ABSOLUTE_Q: AppbioAbsoluteQParser,
    Vendor.BECKMAN_VI_CELL_BLU: ViCellBluParser,
    Vendor.BECKMAN_VI_CELL_XR: ViCellXRParser,
    Vendor.NOVABIO_FLEX2: NovaBioFlexParser,
    Vendor.MOLDEV_SOFTMAX_PRO: SoftmaxproParser,
    Vendor.PERKIN_ELMER_ENVISION: PerkinElmerEnvisionParser,
    Vendor.ROCHE_CEDEX_BIOHT: RocheCedexBiohtParser,
}


class ParserFactory:
    def __init__(self) -> None:
        pass

    def create(
        self, vendor_type: VendorType, default_timezone: Optional[tzinfo] = None
    ) -> VendorParser:
        try:
            timestamp_parser = TimestampParser(default_timezone)
            return _VENDOR_TO_PARSER[Vendor(vendor_type)](timestamp_parser)
        except (ValueError, KeyError) as e:
            error = f"Failed to create parser, unregistered vendor: {vendor_type}"
            raise AllotropeConversionError(error) from e


def get_parser(
    vendor_type: VendorType, default_timezone: Optional[tzinfo] = None
) -> VendorParser:
    return PARSER_FACTORY.create(vendor_type, default_timezone=default_timezone)


PARSER_FACTORY: ParserFactory = ParserFactory()



from datetime import tzinfo
import io
from pathlib import Path
from typing import Any, Optional

from allotropy.allotrope.allotrope import serialize_allotrope
from allotropy.parser_factory import PARSER_FACTORY, VendorType


def allotrope_from_io(
    contents: io.IOBase,
    filename: str,
    vendor_type: VendorType,
    default_timezone: Optional[tzinfo] = None,
) -> dict[str, Any]:
    parser = PARSER_FACTORY.create(vendor_type, default_timezone=default_timezone)
    return serialize_allotrope(parser.to_allotrope(contents, filename))


def allotrope_from_file(
    filepath: str,
    vendor_type: VendorType,
    default_timezone: Optional[tzinfo] = None,
) -> dict[str, Any]:
    with open(filepath, "rb") as f:
        return allotrope_from_io(
            f, Path(filepath).name, vendor_type, default_timezone=default_timezone
        )



from allotropy.allotrope.allotrope import serialize_allotrope
from allotropy.allotrope.models.cell_culture_analyzer_benchling_2023_09_cell_culture_analyzer import (
    AnalyteDocumentItem,
)
from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    DeviceControlAggregateDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleDocument
from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueMillimolePerLiter,
    TQuantityValueNumber,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TDatacube,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
)


def test_basic() -> None:
    model = Model()
    model.measurement_aggregate_document = MeasurementAggregateDocument(
        measurement_identifier="blah",
        plate_well_count=TQuantityValueNumber(1.0),
        measurement_document=[
            MeasurementDocumentItem(
                DeviceControlAggregateDocument(),
                SampleDocument(
                    well_location_identifier="well1", sample_identifier="sample1"
                ),
            )
        ],
    )
    assert serialize_allotrope(model) == {
        "$asm.manifest": "http://purl.allotrope.org/manifests/fluorescence/BENCHLING/2023/09/fluorescence.manifest",
        "measurement aggregate document": {
            "measurement identifier": "blah",
            "plate well count": {
                "value": 1.0,
                "unit": "#",
            },
            "measurement document": [
                {
                    "device control aggregate document": {},
                    "sample document": {
                        "well location identifier": "well1",
                        "sample identifier": "sample1",
                    },
                }
            ],
        },
    }


def test_data_cube() -> None:
    data_cube = TDatacube(
        cube_structure=TDatacubeStructure(
            [
                TDatacubeComponent(
                    FieldComponentDatatype("double"), "elapsed time", "s"
                ),
                TDatacubeComponent(FieldComponentDatatype("int"), "wavelength", None),
            ],
            [
                TDatacubeComponent(
                    FieldComponentDatatype("double"), "fluorescence", "RFU"
                )
            ],
        ),
        data=TDatacubeData(
            [[1.1, 2.2, 3.3], [1.0, 2.0, 3.0]],
            [[4.0, 5.0, None]],
        ),
    )
    assert serialize_allotrope(data_cube) == {
        "cube-structure": {
            "dimensions": [
                {
                    "@componentDatatype": "double",
                    "concept": "elapsed time",
                    "unit": "s",
                },
                {"@componentDatatype": "int", "concept": "wavelength"},
            ],
            "measures": [
                {
                    "@componentDatatype": "double",
                    "concept": "fluorescence",
                    "unit": "RFU",
                }
            ],
        },
        "data": {
            "dimensions": [[1.1, 2.2, 3.3], [1.0, 2.0, 3.0]],
            "measures": [[4.0, 5.0, None]],
        },
    }


def test_omits_null_values_except_for_specified_classes() -> None:
    item = AnalyteDocumentItem(
        "test", TNullableQuantityValueMillimolePerLiter(value=None)
    )

    assert serialize_allotrope(item) == {
        "analyte name": "test",
        "molar concentration": {"unit": "mmol/L", "value": None},
    }



from io import IOBase
from typing import Any, Optional, TypeVar
import uuid

from allotropy.allotrope.allotrope import AllotropyError
from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    ContainerType,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
)
from allotropy.allotrope.models.shared.components.plate_reader import (
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValueNumber,
)
from allotropy.allotrope.models.shared.definitions.definitions import TDateTimeValue
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_structure import (
    Data,
    Plate,
)
from allotropy.parsers.vendor_parser import VendorParser

T = TypeVar("T")


def safe_value(cls: type[T], value: Optional[Any]) -> Optional[T]:
    return None if value is None else cls(value=value)  # type: ignore[call-arg]


class PerkinElmerEnvisionParser(VendorParser):
    def _parse(self, raw_contents: IOBase, _: str) -> Model:
        reader = CsvReader(raw_contents)
        return self._get_model(Data.create(reader))

    def _get_model(self, data: Data) -> Model:
        if not self._check_fluorescence(data):
            msg = "Elmer envision currently only accepts fluorescence data"
            raise NotImplementedError(msg)

        if data.number_of_wells is None:
            msg = "Unable to get number of the wells in the plate"
            raise AllotropyError(msg)

        return Model(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_identifier=str(uuid.uuid4()),
                measurement_time=self._get_measurement_time(data),
                analytical_method_identifier=data.basic_assay_info.protocol_id,
                experimental_data_identifier=data.basic_assay_info.assay_id,
                container_type=ContainerType.well_plate,
                plate_well_count=TQuantityValueNumber(value=data.number_of_wells),
                device_system_document=DeviceSystemDocument(
                    model_number=data.instrument.serial_number,
                    device_identifier=data.instrument.nickname,
                ),
                measurement_document=self._get_measurement_document(data),
            )
        )

    def _check_fluorescence(self, data: Data) -> bool:
        absorbance_patterns = ["ABS", "Absorbance"]
        luminescence_patterns = ["LUM", "Luminescence"]

        for pattern in absorbance_patterns + luminescence_patterns:
            if pattern in data.labels.label:
                return False

        return True

    def _get_measurement_time(self, data: Data) -> TDateTimeValue:
        dates = [
            plate.plate_info.measurement_time
            for plate in data.plates
            if plate.plate_info.measurement_time
        ]

        if dates:
            return self.get_date_time(min(dates))

        msg = "Unable to find valid measurement date"
        raise AllotropyError(msg)

    def _get_device_control_aggregate_document(
        self, data: Data, plate: Plate
    ) -> DeviceControlAggregateDocument:
        ex_filter = data.labels.excitation_filter
        em_filter = data.labels.get_emission_filter(plate.plate_info.emission_filter_id)
        return DeviceControlAggregateDocument(
            [
                DeviceControlDocumentItem(
                    device_type="fluorescence detector",
                    detector_distance_setting__plate_reader_=safe_value(
                        TQuantityValueMillimeter, plate.plate_info.measured_height
                    ),
                    number_of_averages=safe_value(
                        TQuantityValueNumber, data.labels.number_of_flashes
                    ),
                    detector_gain_setting=data.labels.detector_gain_setting,
                    scan_position_setting__plate_reader_=data.labels.scan_position_setting,
                    detector_wavelength_setting=safe_value(
                        TQuantityValueNanometer,
                        em_filter.wavelength if em_filter else None,
                    ),
                    detector_bandwidth_setting=safe_value(
                        TQuantityValueNanometer,
                        em_filter.bandwidth if em_filter else None,
                    ),
                    excitation_wavelength_setting=safe_value(
                        TQuantityValueNanometer,
                        ex_filter.wavelength if ex_filter else None,
                    ),
                    excitation_bandwidth_setting=safe_value(
                        TQuantityValueNanometer,
                        ex_filter.bandwidth if ex_filter else None,
                    ),
                )
            ]
        )

    def _get_measurement_document(self, data: Data) -> list[MeasurementDocumentItem]:
        items = []
        for plate in data.plates:
            if plate.results is None:
                continue

            try:
                p_map = data.plate_maps[plate.plate_info.number]
            except KeyError as e:
                msg = f"Unable to find plate map of {plate.plate_info.barcode}"
                raise AllotropyError(msg) from e

            device_control_aggregate_document = (
                self._get_device_control_aggregate_document(data, plate)
            )

            items += [
                MeasurementDocumentItem(
                    sample_document=SampleDocument(
                        plate_barcode=plate.plate_info.barcode,
                        well_location_identifier=f"{result.col}{result.row}",
                        sample_role_type=p_map.get_sample_role_type(
                            result.col, result.row
                        ),
                    ),
                    device_control_aggregate_document=device_control_aggregate_document,
                    processed_data_aggregate_document=ProcessedDataAggregateDocument(
                        processed_data_document=[
                            ProcessedDataDocumentItem(
                                processed_data=result.value,
                                data_processing_description="processed data",
                            ),
                        ]
                    ),
                    compartment_temperature=safe_value(
                        TQuantityValueDegreeCelsius,
                        plate.plate_info.chamber_temperature_at_start,
                    ),
                )
                for result in plate.results
            ]
        return items



from enum import Enum


class XrVersion(str, Enum):
    _2_06 = "2.06"
    _2_04 = "2.04"


DATE_HEADER = {
    XrVersion._2_06: "Sample date/time",
    XrVersion._2_04: "Sample date",
}

MODEL_RE = r"Vi-CELL XR (?P<version>\d{1,}\.\d{2,}(.\d{1,})?)"
DEFAULT_VERSION = XrVersion._2_06

DEFAULT_ANALYST = "Vi-Cell XR"
MODEL_NUMBER = "Vi-Cell XR"
SOFTWARE_NAME = "Vi-Cell XR"



import io
import itertools
from typing import Any

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parsers.agilent_gen5.create_plate_data import create_plate_data
from allotropy.parsers.agilent_gen5.plate_data import PlateData
from allotropy.parsers.vendor_parser import VendorParser


class AgilentGen5Parser(VendorParser):
    def _get_plate_chunks(self, contents: str) -> list[str]:
        return contents.strip().replace("\r\n", "\n").split(4 * "\n")

    def _parse(self, contents: io.IOBase, filename: str) -> Any:  # noqa: ARG002
        plates: list[PlateData] = []
        software_version_chunk = None
        file_paths_chunk = None
        all_data_chunk = None

        plate_chunks = self._get_plate_chunks(self._read_contents(contents))
        completed_plate = False
        for plate_chunk in plate_chunks:
            if plate_chunk.startswith("Software Version"):
                if completed_plate:
                    new_plate = create_plate_data(
                        software_version_chunk, file_paths_chunk, all_data_chunk
                    )
                    plates.append(new_plate)
                    completed_plate = False
                software_version_chunk = plate_chunk
            elif plate_chunk.startswith("Experiment File Path"):
                file_paths_chunk = plate_chunk
            elif plate_chunk.startswith("Plate Number"):
                all_data_chunk = plate_chunk
                completed_plate = True
            elif plate_chunk.startswith("Actual Temperature"):
                all_data_chunk = f"{all_data_chunk}\n\n{plate_chunk}"

        if completed_plate:
            new_plate = create_plate_data(
                software_version_chunk, file_paths_chunk, all_data_chunk
            )
            plates.append(new_plate)

        if not plates:
            msg = "No plate data found in file"
            raise AllotropeConversionError(msg)

        first_plate = plates[0]
        # TODO we just use the metadata for the first plate, but in theory they could all have
        # different metadata in particular, timestamp is known to be, but not sure on procedures/
        # plate dimensions/etc., need to follow up
        measurement_docs = list(
            itertools.chain.from_iterable([plate.measurement_docs for plate in plates])
        )
        model = first_plate.to_allotrope(measurement_docs)
        model.measurement_aggregate_document.measurement_time = self.get_date_time(
            model.measurement_aggregate_document.measurement_time
        )
        return model

        # TODO stats docs
        # statistics_docs = [plate.statistics_doc for plate in plates]
        # all_statistics_docs = list(itertools.chain.from_iterable(statistics_docs))
        # if all_statistics_docs:
        #     allotrope_dict["measurement aggregate document"]["statistics aggregate document"] = {
        #         "statistics document": all_statistics_docs,
        #     }



from enum import Enum
from typing import Optional


class ReadMode(str, Enum):
    ABSORBANCE = "Absorbance"
    FLUORESCENCE = "Fluorescence"
    LUMINESCENCE = "Luminescence"


class ReadType(str, Enum):
    ENDPOINT = "Endpoint"
    KINETIC = "Kinetic"
    AREASCAN = "Area Scan"
    SPECTRAL = "Spectral"


READTYPE_TO_DIMENSIONS: dict[ReadType, list[tuple[str, str, Optional[str]]]] = {
    ReadType.ENDPOINT: [("int", "wavelength", "nm")],
    ReadType.KINETIC: [("double", "time", "s")],
    ReadType.AREASCAN: [
        ("int", "x", None),
        ("int", "y", None),
    ],
    ReadType.SPECTRAL: [("int", "wavelength", "nm")],
}



from pathlib import Path

from allotropy.allotrope.schema_parser.generate_schemas import generate_schemas


def test_generate_schemas_runs_to_completion() -> None:
    root_dir = Path(__file__).parent.parent.parent.parent
    assert generate_schemas(root_dir) == 9



#!/usr/bin/env python3
from pathlib import Path

from allotropy.allotrope.schema_parser.generate_schemas import generate_schemas

ROOT_DIR = Path(__file__).parent.parent


if __name__ == "__main__":
    generate_schemas(ROOT_DIR)



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import Union

from allotropy.parsers.appbio_quantstudio.referenceable import Referenceable


@dataclass
class DataSource:
    feature: str
    reference: Union[CalculatedDocument, Referenceable]


@dataclass
class CalculatedDocument(Referenceable):
    uuid: str
    name: str
    value: float
    data_sources: list[DataSource]
    iterated: bool = False

    def iter_struct(self) -> Iterator[CalculatedDocument]:
        if self.iterated:
            return

        self.iterated = True
        yield self
        for data_source in self.data_sources:
            if isinstance(data_source.reference, CalculatedDocument):
                yield from data_source.reference.iter_struct()



from datetime import timedelta, timezone, tzinfo
from typing import Optional
from zoneinfo import ZoneInfo

from dateutil import parser
import pytz

from allotropy.allotrope.allotrope import AllotropeConversionError

TIMEZONE_CODES_MAP = {
    **{code: pytz.timezone(code) for code in pytz.all_timezones},
    # Add daylight savings time codes for USA
    **{
        "EDT": timezone(timedelta(hours=-4), "EDT"),
        "CDT": timezone(timedelta(hours=-5), "CDT"),
        "MDT": timezone(timedelta(hours=-6), "MDT"),
        "PDT": timezone(timedelta(hours=-7), "PDT"),
    },
}


class TimestampParser:
    default_timezone: tzinfo

    def __init__(self, default_timezone: Optional[tzinfo] = None):
        if default_timezone and not isinstance(default_timezone, tzinfo):
            msg = f"Invalid default timezone '{default_timezone}'"
            raise AllotropeConversionError(msg)
        self.default_timezone = default_timezone or ZoneInfo("UTC")

    def parse(self, time: Optional[str]) -> Optional[str]:
        if not time:
            return None
        try:
            timestamp = parser.parse(time, tzinfos=TIMEZONE_CODES_MAP, fuzzy=True)
        except ValueError:
            return None
        if not timestamp.tzinfo:
            timestamp = timestamp.replace(tzinfo=self.default_timezone)
        return str(timestamp.isoformat())



from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime
from io import StringIO
from typing import Any, Optional, Union

import numpy as np
import pandas as pd

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TDatacube,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
)
from allotropy.parsers.agilent_gen5.constants import (
    ReadMode,
    ReadType,
    READTYPE_TO_DIMENSIONS,
)
from allotropy.parsers.agilent_gen5.data_point import DataPoint

METADATA_PREFIXES = frozenset(
    {
        "Plate Number",
        "Date",
        "Time",
        "Reader Type:",
        "Reader Serial Number:",
        "Reading Type",
    }
)
GEN5_DATETIME_FORMAT = "%m/%d/%Y %I:%M:%S %p"


def try_float(value: str) -> Union[str, float]:
    try:
        return float(value)
    except ValueError:
        return value


class PlateData(ABC):
    measurements: defaultdict[str, list]
    processed_datas: defaultdict[str, list]
    temperatures: list
    kinetic_times: Optional[list[int]]
    measurement_docs: list
    wells: list
    layout: dict
    concentrations: dict
    read_names: list
    datetime: Optional[str]
    experiment_file_path: Optional[str]
    protocol_file_path: Optional[str]
    read_mode: ReadMode
    read_type: ReadType
    plate_barcode: str
    statistics_doc: list
    actual_temperature: Optional[float]
    data_point_cls: type[DataPoint]

    def __init__(
        self,
        software_version_chunk: str,
        file_paths_chunk: str,
        all_data_chunk: str,
    ):
        self.software_version_chunk = software_version_chunk
        self.file_paths_chunk = file_paths_chunk
        self.all_data_chunk = all_data_chunk

        self.measurements = defaultdict(list)
        self.processed_datas = defaultdict(list)
        self.temperatures = []
        self.kinetic_times = None

        self.measurement_docs = []
        self.wells = []
        self.layout = {}
        self.concentrations = {}
        self.read_names = []
        self.datetime = None
        self.experiment_file_path = None
        self.protocol_file_path = None
        self.statistics_doc = []
        self.actual_temperature = None

        self._parse_software_version_chunk(self.software_version_chunk)
        self._parse_file_paths_chunk(self.file_paths_chunk)
        self._parse_all_data_chunk(self.all_data_chunk)

    def _parse_software_version_chunk(self, software_version_chunk: str) -> None:
        self.software_version = software_version_chunk.split("\t")[1]

    def _parse_file_paths_chunk(self, file_paths_chunk: str) -> None:
        file_paths = file_paths_chunk.split("\n")
        self.experiment_file_path = f"{file_paths[0]}\t".split("\t")[1]
        self.protocol_file_path = f"{file_paths[1]}\t".split("\t")[1]

    def _parse_all_data_chunk(self, all_data_chunk: str) -> None:
        all_data_sections = all_data_chunk.split("\n\n")

        is_kinetic_data = False
        # kinetic_data_label = None
        is_blank_kinetic_data = False
        blank_kinetic_data_label = None

        for data_section in all_data_sections:
            if data_section.startswith("Plate Number"):
                self._parse_metadata(data_section)
            elif data_section.startswith("Plate Type"):
                self._parse_procedure_details(data_section)
            elif data_section.startswith("Layout"):
                self._parse_layout(data_section)
            elif data_section.startswith("Actual Temperature"):
                self._parse_actual_temperature(data_section)
            elif data_section.startswith("Results"):
                self._parse_results(data_section)
            elif data_section.startswith("Curve Name"):
                self._parse_stdcurve(data_section)
            elif is_kinetic_data:
                self._parse_kinetic_data(data_section)
                is_kinetic_data = False
            elif is_blank_kinetic_data:
                self._parse_blank_kinetic_data(data_section, blank_kinetic_data_label)
                is_blank_kinetic_data = False
            elif len(data_section.split("\n")) == 1 and any(
                read_name in data_section for read_name in self.read_names
            ):
                if data_section.startswith("Blank"):
                    is_blank_kinetic_data = True
                    blank_kinetic_data_label = data_section.strip()
                else:
                    is_kinetic_data = True
                    # kinetic_data_label = data_section.strip()

    def _parse_metadata(self, metadata: str) -> None:
        metadata_dict: dict = {}
        metadata_lines = metadata.splitlines()
        for metadata_line in metadata_lines:
            line_split = metadata_line.split("\t")
            if line_split[0] not in METADATA_PREFIXES:
                msg = f"Unrecognized metadata {line_split[0]}"
                raise AllotropeConversionError(msg)
            metadata_dict[line_split[0]] = line_split[1]
        # TODO put more metadata in the right spots

        self.datetime = datetime.strptime(  # noqa: DTZ007
            f"{metadata_dict['Date']} {metadata_dict['Time']}", GEN5_DATETIME_FORMAT
        ).isoformat()

        self.plate_barcode = metadata_dict["Plate Number"]

    def _parse_procedure_details(
        self, procedure_details: str
    ) -> None:  # TODO parse the rest of the procedure details
        if ReadType.KINETIC.value in procedure_details:
            self.read_type = ReadType.KINETIC
        elif ReadType.AREASCAN.value in procedure_details:
            self.read_type = ReadType.AREASCAN
        elif ReadType.SPECTRAL.value in procedure_details:
            self.read_type = ReadType.SPECTRAL
        else:
            self.read_type = (
                ReadType.ENDPOINT
            )  # check for this last, because other modes still contain the word "Endpoint"

        procedure_chunks = self._parse_procedure_chunks(procedure_details)
        for procedure_chunk in procedure_chunks:
            self._parse_procedure_chunk(procedure_chunk)

    def _parse_procedure_chunk(self, procedure_chunk: list[str]) -> None:
        # if no user-defined name is specified for protocols,
        # e.g. it just says "Absorbance Endpoint",
        # Gen5 defaults to using the wavelength as the name
        use_wavelength_names = False
        read_line_length = 2
        wavelength_line_length = 2
        for line in procedure_chunk:
            split_line = line.strip().split("\t")
            if split_line[0] == "Read":
                if len(split_line) != read_line_length:
                    msg = f"Unrecognized Read data {split_line}"
                    raise AllotropeConversionError(msg)
                if split_line[-1] == f"{self.read_mode.title()} Endpoint":
                    use_wavelength_names = True
                else:
                    self.read_names.append(split_line[-1])
            elif split_line[0].startswith("Wavelengths"):
                if use_wavelength_names:
                    split_line_colon = split_line[0].split(":  ")
                    if len(split_line_colon) != wavelength_line_length:
                        msg = f"Unrecognized Wavelengths data {split_line}"
                        raise AllotropeConversionError(msg)
                    self.read_names.extend(split_line_colon[-1].split(", "))

    def _parse_procedure_chunks(self, procedure_details: str) -> list[list[str]]:
        procedure_chunks = []
        current_chunk: list[str] = []
        procedure_lines = procedure_details.splitlines()
        for procedure_line in procedure_lines:
            if procedure_line[0] != "\t":
                procedure_chunks.append(current_chunk)
                current_chunk = []
            current_chunk.append(procedure_line)
        del procedure_chunks[0]
        procedure_chunks.append(current_chunk)

        return procedure_chunks

    def _parse_layout(self, layout: str) -> None:
        layout_lines = layout.splitlines()
        # first line is "Layout", second line is column numbers
        current_row = "A"
        for i in range(2, len(layout_lines)):
            split_line = layout_lines[i].split("\t")
            if split_line[0]:
                current_row = split_line[0]
            label = split_line[-1]
            for j in range(1, len(split_line) - 1):
                well_loc = f"{current_row}{j}"
                if label == "Well ID":
                    self.layout[well_loc] = split_line[j]
                elif label == "Conc/Dil" and split_line[j]:
                    self.concentrations[well_loc] = float(split_line[j])

    def _parse_actual_temperature(self, actual_temperature: str) -> None:
        if len(actual_temperature.split("\n")) != 1:
            msg = f"Unrecognized temperature data {actual_temperature}"
            raise AllotropeConversionError(msg)
        self.actual_temperature = float(actual_temperature.strip().split("\t")[-1])

    def _parse_results(self, results: str) -> None:
        result_lines = results.splitlines()
        if result_lines[0].strip() != "Results":
            msg = f"Unrecognized results data {result_lines[0]}"
            raise AllotropeConversionError(msg)
        # result_lines[1] contains column numbers

        current_row = "A"
        for row_num in range(2, len(result_lines)):
            values = result_lines[row_num].split("\t")
            if values[0]:
                current_row = values[0]
            label = values[-1]  # last column gives information about the type of read
            for col_num in range(1, len(values) - 1):
                well_pos = f"{current_row}{col_num}"
                if well_pos not in self.wells:
                    self.wells.append(well_pos)
                well_value: Union[str, float] = try_float(values[col_num])
                if self._is_processed_data_label(label):
                    self.processed_datas[well_pos].append([label, well_value])
                else:
                    label_only = label.split(":")[-1]
                    self.measurements[well_pos].append([label_only, well_value])

        for well_pos in self.wells:
            datapoint = self.data_point_cls(
                self.read_type,
                self.measurements[well_pos],
                well_pos,
                self.plate_barcode,
                self.layout.get(well_pos),
                self.concentrations.get(well_pos),
                self.processed_datas[well_pos],
                self.actual_temperature,
            )
            self.measurement_docs.append(datapoint.to_measurement_doc())

    def _is_processed_data_label(self, label: str) -> bool:
        if self.read_mode == ReadMode.LUMINESCENCE:
            return not any(
                (label.startswith(read_name) and label.split(":")[-1] == "Lum")
                for read_name in self.read_names
            )
        else:
            return not any(
                (label.startswith(read_name) and label.split(":")[-1][0].isdigit())
                for read_name in self.read_names
            )

    def _parse_stdcurve(self, stdcurve: str) -> None:
        lines = stdcurve.splitlines()
        num_lines = 2
        if len(lines) != num_lines:
            msg = f"Unrecognized std curve data {lines}"
            raise AllotropeConversionError(msg)
        keys = lines[0].split("\t")
        values = lines[1].split("\t")
        self.statistics_doc = [
            {
                "statistical feature": key,
                "feature": try_float(value),
                "group": f"{self.plate_barcode} {self.wells[0]}-{self.wells[-1]}",
            }
            for key, value in zip(keys, values)
        ]

    def _parse_kinetic_data(self, kinetic_data: str) -> None:
        kinetic_data_io = StringIO(kinetic_data)
        df = pd.read_table(kinetic_data_io)
        df_columns = kinetic_data.split("\n")[0].split("\t")
        df = df[
            df["A1"].notna()
        ]  # drop incomplete rows, particularly rows only with "0:00:00"

        self.kinetic_times = [self._hhmmss_to_sec(hhmmss) for hhmmss in df["Time"]]
        self.temperatures = df[df_columns[1]].replace(np.nan, None).tolist()
        has_temperatures = any(temp is not None for temp in self.temperatures)
        for well_pos in df_columns[
            2:
        ]:  # first column is Time, second column is T∞ READ_NAME with no values
            self.wells.append(well_pos)
            values = df[well_pos].tolist()
            if has_temperatures:
                self.measurements[well_pos].extend(
                    list(zip(self.kinetic_times, values, self.temperatures))
                )
            else:
                self.measurements[well_pos].extend(
                    list(zip(self.kinetic_times, values))
                )

    def _parse_blank_kinetic_data(
        self, blank_kinetic_data: str, blank_kinetic_data_label: Optional[str]
    ) -> None:
        blank_kinetic_data_io = StringIO(blank_kinetic_data)
        df = pd.read_table(blank_kinetic_data_io)
        df.dropna(axis=0)  # drop incomplete rows
        df_columns = blank_kinetic_data.split("\n")[0].split("\t")

        for well_pos in df_columns[1:]:  # first column is Time
            measures = df[well_pos].tolist()
            self.processed_datas[well_pos].append(
                [blank_kinetic_data_label, self._blank_data_cube(measures)]
            )

    def _blank_data_cube(self, measures: list[float]) -> TDatacube:
        structure_dimensions = READTYPE_TO_DIMENSIONS[self.read_type]
        structure_measures = [
            ("double", self.read_mode.lower(), self.data_point_cls.unit)
        ]
        return TDatacube(
            label=f"{self.read_type.value.lower()} data",
            cube_structure=TDatacubeStructure(
                [
                    TDatacubeComponent(FieldComponentDatatype(data_type), concept, unit)
                    for data_type, concept, unit in structure_dimensions
                ],
                [
                    TDatacubeComponent(FieldComponentDatatype(data_type), concept, unit)
                    for data_type, concept, unit in structure_measures
                ],
            ),
            data=TDatacubeData([self.kinetic_times], [measures]),  # type: ignore[list-item]
        )

    def _hhmmss_to_sec(
        self, hhmmss: str
    ) -> int:  # convert to seconds, as specified by Allotrope
        hours, minutes, seconds = tuple(int(num) for num in hhmmss.split(":"))
        return (3600 * hours) + (60 * minutes) + seconds

    @abstractmethod
    def to_allotrope(self, measurement_docs: list) -> Any:
        raise NotImplementedError



# plate (repeated for N plates measured)
#     Plate information
#     Background information (optional)
#     Calculated results (optional)
#     Measured results (repeated for N measurements performed per plate)
# Basic assay information
# Protocol information
# Plate type
# Platemap (repeated for N plates measured)
# Calculations
# Auto export parameters
# Operations
# Labels
#     Filters (dependent on detection modality)
#     Mirrors (dependent on detection modality)
# Instrument
from __future__ import annotations

from dataclasses import dataclass
import logging
from re import search
from typing import Optional

import numpy as np
import pandas as pd

from allotropy.allotrope.allotrope import AllotropyError
from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.values import assert_not_none, try_float


def df_to_series(df: pd.DataFrame) -> pd.Series:
    df.columns = df.iloc[0]  # type: ignore[assignment]
    return pd.Series(df.iloc[-1], index=df.columns)


def num_to_chars(n: int) -> str:
    d, m = divmod(n, 26)  # 26 is the number of ASCII letters
    return "" if n < 0 else num_to_chars(d - 1) + chr(m + 65)  # chr(65) = 'A'


@dataclass
class PlateInfo:
    number: str
    barcode: str
    emission_filter_id: str
    measurement_time: Optional[str]
    measured_height: Optional[float]
    chamber_temperature_at_start: Optional[float]

    @staticmethod
    def create(reader: CsvReader) -> Optional[PlateInfo]:
        data = reader.pop_csv_block("^Plate information")
        series = df_to_series(data).replace(np.nan, None)
        series.index = pd.Series(series.index).replace(np.nan, "empty label")  # type: ignore[assignment]

        plate_number = assert_not_none(
            str(series.get("Plate")), "Plate information: Plate"
        )
        barcode = (
            str(series.get("Barcode") or '=""').removeprefix('="').removesuffix('"')
            or f"Plate {plate_number}"
        )

        search_result = search("De=...", str(series.get("Measinfo", "")))
        if not search_result:
            msg = f"Unable to get emition filter id from plate {barcode}"
            raise AllotropyError(msg)
        emission_filter_id = search_result.group().removeprefix("De=")

        measurement_time = str(series.get("Measurement date", ""))

        return PlateInfo(
            plate_number,
            barcode,
            emission_filter_id,
            measurement_time,
            try_float(str(series.get("Measured height"))),
            try_float(str(series.get("Chamber temperature at start"))),
        )


@dataclass
class Result:
    col: str
    row: str
    value: int

    @staticmethod
    def create(reader: CsvReader) -> list[Result]:
        if not reader.current_line_exists() or reader.match(
            "(^Plate information)|(^Basic assay information)"
        ):
            return []

        # Results may or may not have a title
        reader.pop_if_match("^Results")

        data = reader.pop_csv_block()
        series = (
            data.drop(0, axis=0).drop(0, axis=1) if data.iloc[1, 0] == "A" else data
        )
        rows, cols = series.shape
        series.index = [num_to_chars(i) for i in range(rows)]  # type: ignore[assignment]
        series.columns = [str(i).zfill(2) for i in range(1, cols + 1)]  # type: ignore[assignment]

        return [
            Result(col, row, int(series.loc[col, row]))
            for col, row in series.stack().index
        ]


@dataclass
class Plate:
    plate_info: PlateInfo
    results: list[Result]

    @staticmethod
    def create(reader: CsvReader) -> list[Plate]:
        plates: list[Plate] = []

        while True:
            if not reader.match("^Plate information"):
                break

            plate_info: Optional[PlateInfo] = None
            try:
                plate_info = PlateInfo.create(reader)
            except Exception as e:
                logging.warning(f"Failed to parse plate info with error: {e}")

            reader.drop_sections("^Background information|^Calculated results")

            results = Result.create(reader)

            if plate_info:
                plates.append(Plate(plate_info, results=results))

        return plates


@dataclass
class BasicAssayInfo:
    protocol_id: str
    assay_id: str

    @staticmethod
    def create(reader: CsvReader) -> BasicAssayInfo:
        reader.drop_until("^Basic assay information")
        data = reader.pop_csv_block("^Basic assay information").T
        data.iloc[0].replace(":.*", "", regex=True, inplace=True)
        series = df_to_series(data)
        return BasicAssayInfo(
            str(series.get("Protocol ID")),
            str(series.get("Assay ID")),
        )


@dataclass
class PlateType:
    number_of_wells: Optional[float] = None

    @staticmethod
    def create(reader: CsvReader) -> PlateType:
        reader.drop_until("^Plate type")
        data = reader.pop_csv_block("^Plate type").T
        return PlateType(
            try_float(str(df_to_series(data).get("Number of the wells in the plate")))
        )


def get_sample_role_type(encoding: str) -> SampleRoleType:
    # BL        blank               blank_role
    # CTL       control             control_sample_role
    # LB        lance_blank         blank_role
    # LC        lance_crosstalk     control_sample_role
    # LH        lance_high          control_sample_role
    # S         pl_sample           sample_role
    # STD       standard            standard_sample_role
    # -         undefined           <UNDEFINED>
    # UNK       unknown             unknown_sample_role
    # ZH        z_high              control_sample_role
    # ZL        z_low               control_sample_role
    sample_role_type_map = {
        "BL": SampleRoleType.blank_role,
        "CTL": SampleRoleType.control_sample_role,
        "LB": SampleRoleType.blank_role,
        "LC": SampleRoleType.control_sample_role,
        "LH": SampleRoleType.control_sample_role,
        "STD": SampleRoleType.standard_sample_role,
        "S": SampleRoleType.sample_role,
        "-": SampleRoleType.undefined_sample_role,
        "UNK": SampleRoleType.unknown_sample_role,
        "ZH": SampleRoleType.control_sample_role,
        "ZL": SampleRoleType.control_sample_role,
    }
    for pattern, value in sample_role_type_map.items():
        if encoding.startswith(pattern):
            return value

    msg = f"Unable to determine sample role type of plate map encoding {encoding}"
    raise ValueError(msg)


@dataclass
class PlateMap:
    plate_n: str
    group_n: str
    sample_role_type_mapping: dict[str, dict[str, SampleRoleType]]

    @staticmethod
    def create(reader: CsvReader) -> Optional[PlateMap]:
        if not reader.current_line_exists() or reader.match("^Calculations"):
            return None

        plate_n = assert_not_none(reader.pop(), "Platemap number").split(",")[-1]
        group_n = assert_not_none(reader.pop(), "Platemap group").split(",")[-1]

        data = reader.pop_csv_block().replace(" ", "", regex=True)

        reader.pop_data()  # drop type specification
        reader.drop_empty()

        series = (
            data.drop(0, axis=0).drop(0, axis=1) if data.iloc[1, 0] == "A" else data
        )
        rows, cols = series.shape
        series.index = [num_to_chars(i) for i in range(rows)]  # type: ignore[assignment]
        series.columns = [str(i).zfill(2) for i in range(1, cols + 1)]  # type: ignore[assignment]

        sample_role_type_mapping: dict[str, dict[str, SampleRoleType]] = {}
        for row, row_data in series.replace([np.nan, "''"], None).to_dict().items():
            col_mapping: dict[str, SampleRoleType] = {}
            for col, value in row_data.items():
                if value:
                    role_type = get_sample_role_type(str(value))
                    if role_type:
                        col_mapping[str(col)] = role_type
            if col_mapping:
                sample_role_type_mapping[str(row)] = col_mapping

        return PlateMap(plate_n, group_n, sample_role_type_mapping)

    def get_sample_role_type(self, col: str, row: str) -> SampleRoleType:
        try:
            return self.sample_role_type_mapping[row][col]
        except KeyError as e:
            msg = (
                f"Invalid plate map location for plate map {self.plate_n}: {col} {row}"
            )
            raise AllotropyError(msg) from e


def create_plate_maps(reader: CsvReader) -> dict[str, PlateMap]:
    if reader.drop_until("^Platemap") is None:
        msg = "Unable to get plate map information"
        raise AllotropyError(msg)

    reader.pop()  # remove title

    maps: dict[str, PlateMap] = {}
    while _map := PlateMap.create(reader):
        maps[_map.plate_n] = _map
    return maps


@dataclass
class Filter:
    name: str
    wavelength: float
    bandwidth: float

    @staticmethod
    def create(reader: CsvReader) -> Optional[Filter]:
        if not reader.current_line_exists() or reader.match(
            "(^Mirror modules)|(^Instrument:)"
        ):
            return None

        data = reader.pop_csv_block().T
        series = df_to_series(data)

        name = str(series.index[0])

        description = str(series.get("Description"))

        search_result = search("CWL=\\d*nm", description)
        if search_result is None:
            msg = f"Unable to find wavelength for filter {name}"
            raise AllotropyError(msg)
        wavelength = float(
            search_result.group().removeprefix("CWL=").removesuffix("nm")
        )

        search_result = search("BW=\\d*nm", description)
        if search_result is None:
            msg = f"Unable to find bandwidth for filter {name}"
            raise AllotropyError(msg)
        bandwidth = float(search_result.group().removeprefix("BW=").removesuffix("nm"))

        return Filter(name, wavelength, bandwidth)


def create_filters(reader: CsvReader) -> dict[str, Filter]:
    reader.drop_until("(^Filters:)|^Instrument:")

    if reader.match("^Instrument"):
        return {}

    reader.pop()  # remove title

    filters = {}
    while _filter := Filter.create(reader):
        filters[_filter.name] = _filter
    return filters


@dataclass
class Labels:
    label: str
    excitation_filter: Optional[Filter]
    emission_filters: dict[str, Optional[Filter]]
    scan_position_setting: Optional[ScanPositionSettingPlateReader] = None
    number_of_flashes: Optional[float] = None
    detector_gain_setting: Optional[str] = None

    @staticmethod
    def create(reader: CsvReader) -> Labels:
        reader.drop_until("^Labels")
        data = reader.pop_csv_block("^Labels").T
        series = df_to_series(data).replace(np.nan, None)

        filters = create_filters(reader)

        excitation_filter = filters.get(str(series.get("Exc. filter", "")))

        emission_filters = {
            "1st": filters.get(str(series.get("Ems. filter"))),
            "2nd": filters.get(str(series.get("2nd ems. filter"))),
        }
        filter_position_map = {
            "Bottom": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
            "Top": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
        }

        return Labels(
            series.index[0],
            excitation_filter,
            emission_filters,
            filter_position_map.get(str(series.get("Using of emission filter")), None),
            number_of_flashes=try_float(str(series.get("Number of flashes"))),
            detector_gain_setting=str(series.get("Reference AD gain")),
        )

    def get_emission_filter(self, id_val: str) -> Optional[Filter]:
        return self.emission_filters.get(id_val)


@dataclass
class Instrument:
    serial_number: str
    nickname: str

    @staticmethod
    def create(reader: CsvReader) -> Instrument:
        if reader.drop_until("^Instrument") is None:
            msg = "Unable to find instrument information"
            raise AllotropyError(msg)

        reader.pop()  # remove title

        serial_number = assert_not_none(reader.pop(), "serial number").split(",")[-1]
        nickname = assert_not_none(reader.pop(), "nickname").split(",")[-1]

        return Instrument(serial_number, nickname)


@dataclass
class Data:
    plates: list[Plate]
    basic_assay_info: BasicAssayInfo
    number_of_wells: Optional[float]
    plate_maps: dict[str, PlateMap]
    labels: Labels
    instrument: Instrument

    @staticmethod
    def create(reader: CsvReader) -> Data:
        return Data(
            plates=Plate.create(reader),
            basic_assay_info=BasicAssayInfo.create(reader),
            number_of_wells=PlateType.create(reader).number_of_wells,
            plate_maps=create_plate_maps(reader),
            labels=Labels.create(reader),
            instrument=Instrument.create(reader),
        )



# generated by datamodel-codegen:
#   filename:  cell-counting.json
#   timestamp: 2023-10-23T16:36:46+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicroliter,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValueMilliSecond,
    TQuantityValueNanometer,
    TQuantityValuePercent,
    TQuantityValueTODO,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDatacubeData,
    TDatacubeStructure,
    TDateTimeValue,
    TStringValue,
)


@dataclass
class DeviceSystemDocument:
    device_identifier: Optional[TStringValue] = None
    model_number: Optional[TStringValue] = None
    device_serial_number: Optional[TStringValue] = None


@dataclass
class DataSystemDocument:
    data_system_instance_identifier: Optional[TStringValue] = None
    file_name: Optional[TStringValue] = None
    UNC_path: Optional[TStringValue] = None
    software_name: Optional[TStringValue] = None
    software_version: Optional[TStringValue] = None
    ASM_converter_name: Optional[TStringValue] = None
    ASM_converter_version: Optional[TStringValue] = None


@dataclass
class SampleDocument:
    sample_identifier: Optional[TStringValue] = None
    batch_identifier: Optional[TStringValue] = None
    sample_role_type: Optional[TStringValue] = None
    location_identifier: Optional[TStringValue] = None


@dataclass
class ImageDocumentItem:
    experimental_data_identifier: Optional[TStringValue] = None
    field_index: Optional[int] = None


@dataclass
class ImageAggregateDocument:
    image_document: Optional[list[ImageDocumentItem]] = None


@dataclass
class DeviceControlDocumentItem:
    device_type: TStringValue
    sample_volume_setting: Optional[TQuantityValueMicroliter] = None
    detection_type: Optional[TStringValue] = None
    excitation_wavelength_setting: Optional[TQuantityValueNanometer] = None
    emission_wavelength_setting: Optional[TQuantityValueNanometer] = None
    illumination_setting: Optional[TQuantityValuePercent] = None
    exposure_duration_setting: Optional[TQuantityValueMilliSecond] = None
    detector_gain_setting: Optional[TStringValue] = None
    fluorescence_tag_setting: Optional[TStringValue] = None
    field_index: Optional[int] = None


@dataclass
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass
class DataProcessingDocument:
    cell_type_processing_method: Optional[TStringValue] = None
    minimum_cell_diameter: Optional[TQuantityValueMicrometer] = None
    maximum_cell_diameter: Optional[TQuantityValueMicrometer] = None
    cell_density_dilution_factor: Optional[TQuantityValueUnitless] = None


@dataclass
class TotalCellDiameterDistributionDataCube:
    label: Optional[str] = None
    cube_structure: Optional[TDatacubeStructure] = None
    data: Optional[TDatacubeData] = None


@dataclass
class ProcessedDataDocument:
    data_processing_document: DataProcessingDocument
    viability__cell_counter_: TQuantityValuePercent
    viable_cell_density__cell_counter_: TQuantityValueMillionCellsPerMilliliter
    total_cell_count: TQuantityValueCell
    total_cell_density__cell_counter_: Optional[
        TQuantityValueMillionCellsPerMilliliter
    ] = None
    dead_cell_density__cell_counter_: Optional[
        TQuantityValueMillionCellsPerMilliliter
    ] = None
    average_total_cell_diameter: Optional[TQuantityValueMicrometer] = None
    average_live_cell_diameter__cell_counter_: Optional[TQuantityValueMicrometer] = None
    average_dead_cell_diameter__cell_counter_: Optional[TQuantityValueMicrometer] = None
    total_cell_diameter_distribution_data_cube: Optional[
        TotalCellDiameterDistributionDataCube
    ] = None
    viable_cell_count: Optional[TQuantityValueCell] = None
    dead_cell_count: Optional[TQuantityValueCell] = None
    average_total_cell_circularity: Optional[TQuantityValueUnitless] = None
    average_viable_cell_circularity: Optional[TQuantityValueUnitless] = None
    fluorescence_tag_positive_cell_count: Optional[TQuantityValueTODO] = None
    fluorescence_tag_positive_cell_density: Optional[TQuantityValueTODO] = None
    fluorescence_tag_positive_cell_percentage: Optional[TQuantityValueTODO] = None


@dataclass
class MeasurementDocumentItem:
    measurement_identifier: TStringValue
    measurement_time: TDateTimeValue
    sample_document: SampleDocument
    device_control_aggregate_document: DeviceControlAggregateDocument
    processed_data_document: ProcessedDataDocument
    image_aggregate_document: Optional[ImageAggregateDocument] = None


@dataclass
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocumentItem]


@dataclass
class CellCountingDocumentItem:
    analyst: TStringValue
    measurement_aggregate_document: MeasurementAggregateDocument
    submitter: Optional[TStringValue] = None


@dataclass
class CellCountingAggregateDocument:
    device_system_document: DeviceSystemDocument
    cell_counting_document: list[CellCountingDocumentItem]
    data_system_document: Optional[DataSystemDocument] = None


@dataclass
class Model:
    cell_counting_aggregate_document: CellCountingAggregateDocument
    manifest: str = "http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/09/cell-counting.manifest"



import json
import os
from pathlib import Path
from typing import Any

SCHEMAS_DIR = os.path.join(Path(__file__).parent, "schemas")


def add_definitions(schema: dict[str, Any]) -> dict[str, Any]:
    for file, section in [
        ("definitions", "defs"),
        ("units", "defs"),
        ("custom", "custom"),
    ]:
        existing = schema.get(f"${section}", {})
        with open(
            os.path.join(SCHEMAS_DIR, "shared", "definitions", f"{file}.json")
        ) as f:
            additional = json.load(f)
        additional.update(existing)
        schema[f"${section}"] = additional
    return schema


def get_schema(schema_relative_path: str) -> dict[str, Any]:
    with open(os.path.join(SCHEMAS_DIR, schema_relative_path)) as f:
        return add_definitions(json.load(f))



import pytest

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parser_factory import Vendor
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

output_files = (
    "v2.04/Beckman_Vi-Cell-XR_example03_instrumentOutput.xls",
    "v2.06/Beckman_Vi-Cell-XR_example01_instrumentOutput.xlsx",
    "v2.06/Beckman_Vi-Cell-XR_example04_instrumentOutput.xlsx",
    "v2.06/Beckman_Vi-Cell-XR_example05_instrumentOutput.xlsx",
    "v2.06/Beckman_Vi-Cell-XR_example06_instrumentOutput.xlsx",
)

VENDOR_TYPE = Vendor.BECKMAN_VI_CELL_XR
SCHEMA_FILE = "cell-counting/BENCHLING/2023/09/cell-counting.json"


@pytest.mark.parametrize("output_file", output_files)
def test_parse_vi_cell_xr_to_asm_schema_is_valid(output_file: str) -> None:
    test_filepath = f"tests/parsers/beckman_vi_cell_xr/testdata/{output_file}"
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)

    validate_schema(allotrope_dict, SCHEMA_FILE)


@pytest.mark.parametrize("output_file", output_files)
def test_parse_vi_cell_xr_to_asm_expected_contents(output_file: str) -> None:
    test_filepath = f"tests/parsers/beckman_vi_cell_xr/testdata/{output_file}"
    target_filename = output_file.replace(".xlsx", ".json").replace(".xls", ".json")
    expected_filepath = f"tests/parsers/beckman_vi_cell_xr/testdata/{target_filename}"
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)

    validate_contents(allotrope_dict, expected_filepath)


def test_perse_vi_cell_xr_file_without_required_fields_then_raise() -> None:
    test_filepath = "tests/parsers/beckman_vi_cell_xr/testdata/v2.04/Beckman_Vi-Cell-XR_example02_instrumentOutput.xls"
    with pytest.raises(AllotropeConversionError):
        from_file(test_filepath, VENDOR_TYPE)



from dataclasses import dataclass
from typing import Optional


@dataclass
class Centimeter:
    unit: Optional[str] = "cm"


@dataclass
class Millimeter:
    unit: Optional[str] = "mm"


@dataclass
class Nanometer:
    unit: Optional[str] = "nm"


@dataclass
class Micrometer:
    unit: Optional[str] = "μm"


@dataclass
class Hertz:
    unit: Optional[str] = "Hz"


@dataclass
class CubicMillimeter:
    unit: Optional[str] = "mm^3"


@dataclass
class Unitless:
    unit: Optional[str] = "(unitless)"


@dataclass
class SecondTime:
    unit: Optional[str] = "s"


@dataclass
class Percent:
    unit: Optional[str] = "%"


@dataclass
class Cell:
    unit: Optional[str] = "cell"


@dataclass
class Microliter:
    unit: Optional[str] = "μL"


@dataclass
class NumberPerMicroliter:
    unit: Optional[str] = "#/μL"


@dataclass
class MilliSecond:
    unit: Optional[str] = "ms"


@dataclass
class MillionCellsPerMilliliter:
    unit: Optional[str] = "10^6 cells/mL"


@dataclass
class TODO:
    unit: Optional[str] = "TODO"


@dataclass
class DegreeCelsius:
    unit: Optional[str] = "degC"


@dataclass
class Number:
    unit: Optional[str] = "#"


@dataclass
class PicogramPerMilliliter:
    unit: Optional[str] = "pg/mL"


@dataclass
class GramPerLiter:
    unit: Optional[str] = "g/L"


@dataclass
class UnitPerLiter:
    unit: Optional[str] = "U/L"


@dataclass
class MillimeterOfMercury:
    unit: Optional[str] = "mmHg"


@dataclass
class OpticalDensity:
    unit: Optional[str] = "OD"


@dataclass
class PH:
    unit: Optional[str] = "pH"


@dataclass
class MilliOsmolesPerKilogram:
    unit: Optional[str] = "mosm/kg"


@dataclass
class MillimolePerLiter:
    unit: Optional[str] = "mmol/L"


@dataclass
class MilliAbsorbanceUnit:
    unit: Optional[str] = "mAU"


@dataclass
class RelativeFluorescenceUnit:
    unit: Optional[str] = "RFU"


@dataclass
class RelativeLightUnit:
    unit: Optional[str] = "RLU"



from __future__ import annotations

from dataclasses import fields, is_dataclass
from typing import Any, Callable, cast

import cattrs
from cattrs.gen import make_dict_unstructure_fn, override

from allotropy.allotrope.models.cell_culture_analyzer_benchling_2023_09_cell_culture_analyzer import (
    AnalyteDocumentItem,
    MeasurementDocumentItem,
)
from allotropy.allotrope.models.pcr_benchling_2023_09_qpcr import (
    ProcessedDataDocumentItem,
)


class AllotropeConversionError(Exception):
    pass


class AllotropyError(Exception):
    pass


# TODO: gather exceptions when parsing models from schema and publish them in model
SPECIAL_KEYS = {
    "manifest": "$asm.manifest",
    "field_asm_manifest": "$asm.manifest",
    "cube_structure": "cube-structure",
    "field_componentDatatype": "@componentDatatype",
    "field_asm_fill_value": "$asm.fill-value",
    "field_type": "@type",
    "field_index": "@index",
    "scan_position_setting__plate_reader_": "scan position setting (plate reader)",
    "detector_distance_setting__plate_reader_": "detector distance setting (plate reader)",
    "cell_type__cell_counter_": "cell type (cell counter)",
    "dead_cell_density__cell_counter_": "dead cell density (cell counter)",
    "average_dead_cell_diameter__cell_counter_": "average dead cell diameter (cell counter)",
    "viability__cell_counter_": "viability (cell counter)",
    "total_cell_density__cell_counter_": "total cell density (cell counter)",
    "viable_cell_density__cell_counter_": "viable cell density (cell counter)",
    "average_live_cell_diameter__cell_counter_": "average live cell diameter (cell counter)",
    "average_total_cell_diameter__cell_counter_": "average total cell diameter (cell counter)",
    "total_cell_diameter_distribution__cell_counter_": "total cell diameter distribution (cell counter)",
    "viable_cell_count__cell_counter_": "viable cell count (cell counter)",
    "total_cell_count__cell_counter_": "total cell count (cell counter)",
    "pco2": "pCO2",
    "co2_saturation": "CO2 saturation",
    "po2": "pO2",
    "o2_saturation": "O2 saturation",
}


CELL_CULTURE_NULLABLE_VALUE_CLASSES: dict[Any, set[str]] = {
    MeasurementDocumentItem: {
        "pco2",
        "co2_saturation",
        "po2",
        "o2_saturation",
        "optical_density",
        "pH",
        "osmolality",
        "viability__cell_counter_",
        "total_cell_density__cell_counter_",
        "viable_cell_density__cell_counter_",
        "average_live_cell_diameter__cell_counter_",
        "average_total_cell_diameter__cell_counter_",
        "total_cell_diameter_distribution__cell_counter_",
        "viable_cell_count__cell_counter_",
        "total_cell_count__cell_counter_",
    },
    AnalyteDocumentItem: {"molar_concentration"},
}

QPCR_NULLABLE_VALUE_CLASSES: dict[Any, set[str]] = {
    ProcessedDataDocumentItem: {"cycle_threshold_result"}
}

EMPTY_VALUE_CLASS_AND_FIELD = {
    **CELL_CULTURE_NULLABLE_VALUE_CLASSES,
    **QPCR_NULLABLE_VALUE_CLASSES,
}


def should_allow_empty_value_field(cls: Any, key: str) -> bool:
    return key in EMPTY_VALUE_CLASS_AND_FIELD.get(cls, set())


def get_key(key: str) -> str:
    return SPECIAL_KEYS.get(key, key.replace("_", " "))


def serialize_allotrope(model: Any) -> dict[str, Any]:
    converter = cattrs.Converter()

    # Default check for omitting values skips if value is None
    def should_omit(_: str, v: Any) -> bool:
        return v is None

    # Special should_omit check for allowing an empty value for 'value' keys, controlled by should_allow_empty_value_field
    def should_omit_allow_empty_value_field(k: str, v: Any) -> bool:
        return v is None and k != "value"

    unstructure_fn_cache = {}

    def unstructure_dataclass_fn(
        cls: Any, should_omit: Callable[[str, Any], bool] = should_omit
    ) -> Callable[[Any], dict]:
        def unstructure(obj: Any) -> Any:
            # Break out of dataclass recursion by calling back to converter.unstructure
            if not is_dataclass(obj):
                return converter.unstructure(obj)

            return {
                get_key(k): v
                for k, v in make_unstructure_fn(type(obj))(obj).items()
                if not should_omit(k, v)
            }

        # This custom unstructure function overrides the unstruct_hook when we should should_allow_empty_value_field.
        # We need to do this at this level because we need to know both the parent class and the field name at the
        # same time.
        def make_unstructure_fn(subcls: Any) -> Callable[[Any], dict[str, Any]]:
            if (cls, subcls) not in unstructure_fn_cache:
                unstructure_fn_cache[(cls, subcls)] = make_dict_unstructure_fn(
                    subcls,
                    converter,
                    **{
                        a.name: override(
                            unstruct_hook=unstructure_dataclass_fn(
                                subcls, should_omit_allow_empty_value_field
                            )
                            if should_allow_empty_value_field(cls, a.name)
                            else None
                        )
                        for a in fields(cls)
                    },
                )
            return unstructure_fn_cache[(cls, subcls)]

        return unstructure

    converter.register_unstructure_hook_factory(is_dataclass, unstructure_dataclass_fn)
    result = converter.unstructure(model)
    return cast(dict[str, Any], result)



from typing import Any, Optional
import uuid

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueNanometer,
    TQuantityValueNumber,
)
from allotropy.allotrope.models.ultraviolet_absorbance_benchling_2023_09_ultraviolet_absorbance import (
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
)
from allotropy.parsers.moldev_softmax_pro.plate_block import PlateBlock, WellData
from allotropy.parsers.utils.values import assert_not_none, natural_sort_key, try_int


class AbsorbancePlateBlock(PlateBlock):
    READ_MODE = "Absorbance"
    CONCEPT = "absorbance"
    UNIT = "mAU"
    DATA_TYPE_IDX = 6

    def parse_read_mode_header(self, header: list[Optional[str]]) -> None:
        [self.first_row, num_rows] = header[19:21]
        self.pmt_gain = None
        self.num_rows = assert_not_none(try_int(num_rows), "num_rows")

    def generate_device_control_doc(self) -> DeviceControlDocumentItem:
        device_control_doc = DeviceControlDocumentItem(
            detector_gain_setting=self.pmt_gain
        )

        if self.is_single_wavelength:
            device_control_doc.detector_wavelength_setting = TQuantityValueNanometer(
                self.wavelengths[0]
            )

        return device_control_doc

    def generate_measurement_doc(
        self, well: str, well_data: WellData
    ) -> MeasurementDocumentItem:
        measurement = MeasurementDocumentItem(
            DeviceControlAggregateDocument([self.generate_device_control_doc()]),
            self.generate_sample_document(well),
        )

        if well_data.temperature is not None:
            measurement.compartment_temperature = TQuantityValueDegreeCelsius(
                float(well_data.temperature)
            )

        if not well_data.is_empty:
            measurement.data_cube = self.generate_data_cube(well_data)

        if well_data.processed_data:
            measurement.processed_data_aggregate_document = (
                self.generate_processed_data_aggreate_document(well_data)
            )

        return measurement

    def to_allotrope(self) -> Any:
        wells = sorted(self.well_data.keys(), key=natural_sort_key)

        allotrope_file = Model(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_identifier=str(uuid.uuid4()),
                plate_well_count=TQuantityValueNumber(self.num_wells),
                measurement_document=[
                    self.generate_measurement_doc(well, self.well_data[well])
                    for well in wells
                ],
            )
        )

        return allotrope_file



from io import IOBase

import numpy as np
import pandas as pd


class AbsoluteQReader:
    def __init__(self, contents: IOBase):
        absolute_q_data = pd.read_csv(  # type: ignore[call-overload]
            filepath_or_buffer=contents, parse_dates=["Date"]
        ).replace(np.nan, None)
        self.wells: pd.DataFrame = absolute_q_data.dropna(subset=["Name"])
        self.group_rows: pd.DataFrame = absolute_q_data[absolute_q_data["Name"].isna()]



from __future__ import annotations

from collections.abc import Iterator
from typing import Generic, Optional, TypeVar, Union

from allotropy.allotrope.allotrope import AllotropeConversionError

T = TypeVar("T")


class ViewData(Generic[T]):
    def __init__(self, data: dict[str, Union[ViewData, list[T]]]):
        self.data = data

    def get_first_key(self) -> str:
        key, *_ = self.data.keys()
        return key

    def iter_keys(self) -> Iterator[list[str]]:
        for key, item in self.data.items():
            if isinstance(item, ViewData):
                for sub_keys in item.iter_keys():
                    yield [key, *sub_keys]
            else:
                yield [key]

    def get_item(self, *keys: str) -> Union[ViewData, list[T]]:
        if len(keys) == 0:
            return self.data[self.get_first_key()]

        key, *sub_keys = keys
        item = self.data[key]
        if isinstance(item, ViewData):
            return item.get_item(*sub_keys)
        else:
            return item

    def get_sub_view_data(self, *keys: str) -> ViewData:
        item = self.get_item(*keys)
        if isinstance(item, ViewData):
            return item
        msg = f"Unable to get sub view data with keys {keys}"
        raise AllotropeConversionError(msg)

    def get_leaf_item(self, *keys: str) -> list[T]:
        item = self.get_item(*keys)
        if isinstance(item, ViewData):
            msg = f"Unable to get leaf item of view data with keys {keys}"
            raise AllotropeConversionError(msg)
        return item


class View(Generic[T]):
    def __init__(self, sub_view: Optional[View] = None):
        self.sub_view = sub_view

    def sort_elements(self, _: list[T]) -> dict[str, list[T]]:
        return {}

    def apply(self, elements: list[T]) -> ViewData:
        return ViewData(
            {
                id_: self.sub_view.apply(element) if self.sub_view else element
                for id_, element in self.sort_elements(elements).items()
            }
        )



from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    ContainerType,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.models.shared.components.plate_reader import (
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SampleRoleType,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValueNumber,
)
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_structure import (
    BasicAssayInfo,
    Data,
    Filter,
    Instrument,
    Labels,
    Plate,
    PlateInfo,
    PlateMap,
    Result,
)


def get_data() -> Data:
    return Data(
        plates=[
            Plate(
                plate_info=PlateInfo(
                    number="1",
                    barcode="Plate 1",
                    emission_filter_id="1st",
                    measurement_time="10/13/2022 3:08:06 PM",
                    measured_height=11.9,
                    chamber_temperature_at_start=23.17,
                ),
                results=[Result(col="A", row="01", value=31441)],
            ),
            Plate(
                plate_info=PlateInfo(
                    number="1",
                    barcode="Plate 1",
                    emission_filter_id="2nd",
                    measurement_time="10/13/2022 3:08:06 PM",
                    measured_height=11.9,
                    chamber_temperature_at_start=23.17,
                ),
                results=[Result(col="A", row="01", value=80368)],
            ),
        ],
        basic_assay_info=BasicAssayInfo("100302", "3134"),
        number_of_wells=96.0,
        plate_maps={
            "1": PlateMap(
                plate_n="1",
                group_n="1",
                sample_role_type_mapping={
                    "01": {"A": SampleRoleType.undefined_sample_role}
                },
            ),
        },
        labels=Labels(
            label="AC HTRF Laser [Eu]",
            excitation_filter=Filter("UV2 (TRF) 320", 320, 75),
            emission_filters={
                "1st": Filter("APC 665", 665, 75),
                "2nd": Filter("Cy5 620", 620, 10),
            },
            scan_position_setting=ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
            number_of_flashes=50,
            detector_gain_setting="2",
        ),
        instrument=Instrument(
            serial_number="1050209",
            nickname="EnVision",
        ),
    )


def get_model() -> Model:
    return Model(
        manifest="http://purl.allotrope.org/manifests/fluorescence/BENCHLING/2023/09/fluorescence.manifest",
        measurement_aggregate_document=MeasurementAggregateDocument(
            measurement_identifier="",
            measurement_time="2022-10-13T15:08:06+00:00",
            analytical_method_identifier="100302",
            experimental_data_identifier="3134",
            container_type=ContainerType.well_plate,
            plate_well_count=TQuantityValueNumber(value=96.0),
            device_system_document=DeviceSystemDocument(
                model_number="1050209",
                device_identifier="EnVision",
            ),
            measurement_document=[
                MeasurementDocumentItem(
                    device_control_aggregate_document=DeviceControlAggregateDocument(
                        device_control_document=[
                            DeviceControlDocumentItem(
                                device_type="fluorescence detector",
                                shaking_configuration_description=None,
                                detector_distance_setting__plate_reader_=TQuantityValueMillimeter(
                                    value=11.9,
                                    has_statistic_datum_role=None,
                                    unit="mm",
                                    field_type=None,
                                ),
                                integration_time=None,
                                number_of_averages=TQuantityValueNumber(
                                    value=50.0,
                                    has_statistic_datum_role=None,
                                    unit="#",
                                    field_type=None,
                                ),
                                detector_gain_setting="2",
                                scan_position_setting__plate_reader_=ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
                                detector_carriage_speed_setting=None,
                                detector_wavelength_setting=TQuantityValueNanometer(
                                    value=665.0,
                                    has_statistic_datum_role=None,
                                    unit="nm",
                                    field_type=None,
                                ),
                                detector_bandwidth_setting=TQuantityValueNanometer(
                                    value=75.0,
                                    has_statistic_datum_role=None,
                                    unit="nm",
                                    field_type=None,
                                ),
                                excitation_bandwidth_setting=TQuantityValueNanometer(
                                    value=75.0,
                                    has_statistic_datum_role=None,
                                    unit="nm",
                                    field_type=None,
                                ),
                                excitation_wavelength_setting=TQuantityValueNanometer(
                                    value=320.0,
                                    has_statistic_datum_role=None,
                                    unit="nm",
                                    field_type=None,
                                ),
                                wavelength_filter_cutoff_setting=None,
                                field_index=None,
                            )
                        ]
                    ),
                    sample_document=SampleDocument(
                        well_location_identifier="A01",
                        sample_identifier=None,
                        batch_identifier=None,
                        sample_role_type=SampleRoleType.undefined_sample_role,
                        plate_barcode="Plate 1",
                    ),
                    data_cube=None,
                    compartment_temperature=TQuantityValueDegreeCelsius(
                        value=23.17,
                        has_statistic_datum_role=None,
                        unit="degC",
                        field_type=None,
                    ),
                    processed_data_aggregate_document=ProcessedDataAggregateDocument(
                        processed_data_document=[
                            ProcessedDataDocumentItem(
                                processed_data=31441,
                                data_format_specification_type=None,
                                data_processing_description="processed data",
                            )
                        ]
                    ),
                    mass_concentration=None,
                ),
                MeasurementDocumentItem(
                    device_control_aggregate_document=DeviceControlAggregateDocument(
                        device_control_document=[
                            DeviceControlDocumentItem(
                                device_type="fluorescence detector",
                                shaking_configuration_description=None,
                                detector_distance_setting__plate_reader_=TQuantityValueMillimeter(
                                    value=11.9,
                                    has_statistic_datum_role=None,
                                    unit="mm",
                                    field_type=None,
                                ),
                                integration_time=None,
                                number_of_averages=TQuantityValueNumber(
                                    value=50.0,
                                    has_statistic_datum_role=None,
                                    unit="#",
                                    field_type=None,
                                ),
                                detector_gain_setting="2",
                                scan_position_setting__plate_reader_=ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
                                detector_carriage_speed_setting=None,
                                detector_wavelength_setting=TQuantityValueNanometer(
                                    value=620.0,
                                    has_statistic_datum_role=None,
                                    unit="nm",
                                    field_type=None,
                                ),
                                detector_bandwidth_setting=TQuantityValueNanometer(
                                    value=10.0,
                                    has_statistic_datum_role=None,
                                    unit="nm",
                                    field_type=None,
                                ),
                                excitation_bandwidth_setting=TQuantityValueNanometer(
                                    value=75.0,
                                    has_statistic_datum_role=None,
                                    unit="nm",
                                    field_type=None,
                                ),
                                excitation_wavelength_setting=TQuantityValueNanometer(
                                    value=320.0,
                                    has_statistic_datum_role=None,
                                    unit="nm",
                                    field_type=None,
                                ),
                                wavelength_filter_cutoff_setting=None,
                                field_index=None,
                            )
                        ]
                    ),
                    sample_document=SampleDocument(
                        well_location_identifier="A01",
                        sample_identifier=None,
                        batch_identifier=None,
                        sample_role_type=SampleRoleType.undefined_sample_role,
                        plate_barcode="Plate 1",
                    ),
                    data_cube=None,
                    compartment_temperature=TQuantityValueDegreeCelsius(
                        value=23.17,
                        has_statistic_datum_role=None,
                        unit="degC",
                        field_type=None,
                    ),
                    processed_data_aggregate_document=ProcessedDataAggregateDocument(
                        processed_data_document=[
                            ProcessedDataDocumentItem(
                                processed_data=80368,
                                data_format_specification_type=None,
                                data_processing_description="processed data",
                            )
                        ]
                    ),
                    mass_concentration=None,
                ),
            ],
        ),
    )



import pytest

from allotropy.allotrope.models.shared.definitions.definitions import TDatacubeData


def test_data_cube_data() -> None:
    cube = TDatacubeData(
        dimensions=[["A", "B"]],
        measures=[[1.0, None]],
    )
    assert cube.dimensions == [["A", "B"]]
    assert cube.measures == [[1.0, None]]

    cube = TDatacubeData(
        dimensions=[[1.0, 2.0]],
        points=[[True, False]],
    )
    assert cube.dimensions == [[1.0, 2.0]]
    assert cube.points == [[True, False]]


def test_data_cube_data_oneof_post_init() -> None:
    with pytest.raises(
        ValueError, match="Exactly one of measures or points must be set"
    ):
        TDatacubeData(dimensions=[["A", "B"]])

    with pytest.raises(
        ValueError, match="Exactly one of measures or points must be set"
    ):
        TDatacubeData(dimensions=[["A", "B"]], measures=[], points=[])


def test_data_cube_data_oneof_post_init_empty_list_ok() -> None:
    cube = TDatacubeData(dimensions=[["A", "B"]], measures=[])
    assert cube.measures == []
    assert cube.points is None

    cube = TDatacubeData(dimensions=[["A", "B"]], points=[])
    assert cube.measures is None
    assert cube.points == []



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

TTupleData = list[Optional[Union[float, bool, str]]]


TUnit = str


TBooleanArray = list[bool]


TBooleanOrNullArray = list[Optional[bool]]


TNumberArray = list[float]


TNumberOrNullArray = list[Optional[float]]


TStringArray = list[str]


TStringOrNullArray = list[Optional[str]]


TClass = str


@dataclass
class TStringValueItem:
    value: str
    field_type: str


TStringValue = Union[str, TStringValueItem]


@dataclass
class TDateTimeValueItem:
    field_type: TClass
    value: str


TDateTimeValue = Union[str, TDateTimeValueItem]
TDateTimeStampValue = TDateTimeValue


@dataclass
class TBooleanValueItem:
    field_type: TClass
    value: bool


TBooleanValue = Union[bool, TBooleanValueItem]


class TStatisticDatumRole(Enum):
    arithmetic_mean_role = "arithmetic mean role"
    median_role = "median role"
    relative_standard_deviation_role = "relative standard deviation role"
    skewness_role = "skewness role"
    standard_deviation_role = "standard deviation role"
    variance_role = "variance role"
    maximum_value_role = "maximum value role"
    minimum_value_role = "minimum value role"


@dataclass
class TQuantityValue:
    value: float
    unit: TUnit
    has_statistic_datum_role: Optional[TStatisticDatumRole] = None
    field_type: Optional[TClass] = None


@dataclass
class TNullableQuantityValue:
    value: Optional[float]
    unit: TUnit
    has_statistic_datum_role: Optional[TStatisticDatumRole] = None
    field_type: Optional[TClass] = None


# NOTE: this is defined to allow override of unit default for TQuaniityValue<Unit> (otherwise mypy gets mad)
@dataclass
class TQuantityValueWithOptionalUnit:
    value: float
    unit: Optional[TUnit]
    has_statistic_datum_role: Optional[TStatisticDatumRole] = None
    field_type: Optional[TClass] = None


@dataclass
class TNullableQuantityValueWithOptionalUnit:
    value: Optional[float]
    unit: Optional[TUnit]
    has_statistic_datum_role: Optional[TStatisticDatumRole] = None
    field_type: Optional[TClass] = None


class FieldComponentDatatype(Enum):
    double = "double"
    float = "float"  # noqa: A003
    decimal = "decimal"
    integer = "integer"
    byte = "byte"
    int = "int"  # noqa: A003
    short = "short"
    long = "long"
    string = "string"
    boolean = "boolean"
    dateTime = "dateTime"


class Scale(Enum):
    nominal = "nominal"
    ordinal = "ordinal"
    cardinal = "cardinal"
    interval = "interval"
    range = "range"  # noqa: A003


class Type(Enum):
    linear = "linear"
    logarithmic = "logarithmic"


@dataclass
class TFunction:
    type: Optional[Type] = Type.linear  # noqa: A003
    start: Optional[float] = 1
    length: Optional[float] = None
    incr: Optional[float] = 1


@dataclass
class TDatacubeComponent:
    field_componentDatatype: FieldComponentDatatype
    concept: TClass
    unit: Optional[TUnit] = None
    scale: Optional[Scale] = None
    field_asm_fill_value: Optional[Union[str, float, int, bool]] = None


TDimensionArray = Union[TNumberArray, TBooleanArray, TStringArray]


TMeasureArray = Union[TNumberOrNullArray, TBooleanOrNullArray, TStringOrNullArray]


@dataclass
class TDatacubeStructure:
    dimensions: list[TDatacubeComponent]
    measures: list[TDatacubeComponent]


"""
TODO: datamodel-codegen cannot properly generate the models for TDatacubeData with the
oneOf{measures, points} constraint. I can't figure out how to do it correctly right now either.
We need to either figure this out and file a bug with datamodel-codegen or fix the schema.

@dataclass
class TDimensionData:
    dimensions: List[Union[TDimensionArray, TFunction]]


@dataclass
class TMeasureDatum:
    measures: List[TMeasureArray]


@dataclass
class TMeasureDatum1:
    points: List[TTupleData]


TMeasureData = Union[TMeasureDatum, TMeasureDatum1]


@dataclass
class TDatacubeData(TDimensionData):
    pass
"""


@dataclass
class TDatacubeData:
    dimensions: list[Union[TDimensionArray, TFunction]]
    measures: Optional[list[TMeasureArray]] = None
    points: Optional[list[TTupleData]] = None

    def __post_init__(self) -> None:
        # Logic for enforcing oneOf
        if not (self.measures is None) ^ (self.points is None):
            error = "Exactly one of measures or points must be set"
            raise ValueError(error)


@dataclass
class TDatacube:
    label: Optional[str] = None
    cube_structure: Optional[TDatacubeStructure] = None
    data: Optional[TDatacubeData] = None



import json
import os
from pathlib import Path
import re
import subprocess  # noqa: S404, RUF100

from autoflake import fix_file  # type: ignore[import-untyped]
from datamodel_code_generator import (
    DataModelType,
    generate,
    InputFileType,
    PythonVersion,
)

from allotropy.allotrope.schema_parser.model_class_editor import modify_file
from allotropy.allotrope.schemas import get_schema

SCHEMA_DIR_PATH = "src/allotropy/allotrope/schemas"
MODEL_DIR_PATH = "src/allotropy/allotrope/models"


def lint_file(model_path: str) -> None:
    # The first run of ruff changes typing annotations and causes unused imports. We catch failure
    # due to unused imports.
    try:
        subprocess.check_call(
            f"ruff {model_path} --fix",
            shell=True,  # noqa: S602
            stdout=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        pass
    # The call to autoflake.fix_file removes unused imports.
    fix_file(
        model_path,
        {
            "in_place": True,
            "remove_unused_variables": True,
            "write_to_stdout": False,
            "ignore_init_module_imports": False,
            "expand_star_imports": False,
            "remove_all_unused_imports": True,
            "remove_duplicate_keys": True,
            "remove_rhs_for_unused_variables": False,
            "ignore_pass_statements": False,
            "ignore_pass_after_docstring": False,
            "check": False,
            "check_diff": False,
        },
    )
    # The second call to ruff checks for additional rules.
    subprocess.check_call(
        f"ruff {model_path} --fix", shell=True, stdout=subprocess.DEVNULL  # noqa: S602
    )
    subprocess.check_call(
        f"black {model_path}", shell=True, stderr=subprocess.DEVNULL  # noqa: S602
    )


def files_equal(path1: str, path2: str) -> bool:
    with open(path1) as file1, open(path2) as file2:
        for line1, line2 in zip(file1, file2):
            if line1 != line2 and not line1.startswith("#   timestamp:"):
                return False
    return True


def generate_schemas(root_dir: Path) -> int:
    """Generate schemas from JSON schema files.

    :root_dir: The root directory of the project.
    :return: The number of schemas generated.
    """

    os.chdir(os.path.join(root_dir, SCHEMA_DIR_PATH))
    schema_paths = list(Path(".").rglob("*.json"))
    os.chdir(os.path.join(root_dir))
    number_generated = 0
    for rel_schema_path in schema_paths:
        if str(rel_schema_path).startswith("shared"):
            continue
        print(f"Generating models for schema: {rel_schema_path}...")  # noqa: T201
        schema_name = rel_schema_path.stem
        schema_path = os.path.join(root_dir, SCHEMA_DIR_PATH, rel_schema_path)
        output_file = re.sub(
            "/|-", "_", f"{rel_schema_path.parent}_{schema_name}.py"
        ).lower()
        model_path = os.path.join(root_dir, MODEL_DIR_PATH, output_file)
        model_backup_path = f"{model_path}.bak"
        schema_backup_path = f"{schema_path}.bak"

        # Backup model to do diff after generation
        if os.path.exists(model_path):
            os.rename(model_path, model_backup_path)

        # Backup schema and override with extra defs
        schema = get_schema(str(rel_schema_path))
        os.rename(schema_path, schema_backup_path)
        with open(schema_path, "w") as f:
            json.dump(schema, f)

        # Generate models
        generate(
            input_=Path(schema_path),
            output=Path(model_path),
            output_model_type=DataModelType.DataclassesDataclass,
            input_file_type=InputFileType.JsonSchema,
            # Specify base_class as empty when using dataclass
            base_class="",
            target_python_version=PythonVersion.PY_39,
            use_union_operator=False,
        )
        # Import classes from shared files, remove unused classes, format.
        modify_file(model_path, schema_path)
        lint_file(model_path)

        # Restore backups
        if os.path.exists(model_backup_path):
            if files_equal(model_path, model_backup_path):
                os.rename(model_backup_path, model_path)
            else:
                os.remove(model_backup_path)
        os.rename(schema_backup_path, schema_path)

        number_generated += 1

    return number_generated



import pandas as pd

from allotropy.allotrope.models.cell_counting_benchling_2023_09_cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ProcessedDataDocument,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION


def get_filename() -> str:
    return "example.csv"


def get_data() -> pd.DataFrame:
    return pd.DataFrame(
        data=[
            [
                "Vi-CELL",
                "TF-B0256-EXPN-SF",
                "6.0",
                "30.0",
                "2022-03-21 16:56:00",
                "Mammalian",
                "1.0",
                "96.9",
                "0.78",
                "0.75",
                "18.51",
                "18.72",
                "2191.0",
                "2123.0",
                "0.88",
                "0.88",
            ]
        ],
        columns=[
            "Analysis by",
            "Sample ID",
            "Minimum Diameter (μm)",
            "Maximum Diameter (μm)",
            "Analysis date/time",
            "Cell type",
            "Dilution",
            "Viability (%)",
            "Total (x10^6) cells/mL",
            "Viable (x10^6) cells/mL",
            "Average diameter (μm)",
            "Average viable diameter (μm)",
            "Cell count",
            "Viable cells",
            "Average circularity",
            "Average viable circularity",
        ],
    )


def get_model() -> Model:
    data = get_data()
    filename = get_filename()
    sample = data.loc[0]
    return Model(
        cell_counting_aggregate_document=CellCountingAggregateDocument(
            device_system_document=DeviceSystemDocument(
                model_number="Vi-Cell BLU",
            ),
            data_system_document=DataSystemDocument(
                file_name=filename,
                software_name="Vi-Cell BLU",
                ASM_converter_name=ASM_CONVERTER_NAME,
                ASM_converter_version=ASM_CONVERTER_VERSION,
            ),
            cell_counting_document=[
                CellCountingDocumentItem(
                    analyst=sample.get("Analysis by"),  # type: ignore[arg-type]
                    measurement_aggregate_document=MeasurementAggregateDocument(
                        measurement_document=[
                            MeasurementDocumentItem(
                                measurement_identifier="",
                                measurement_time="2022-03-21T16:56:00+00:00",
                                sample_document=SampleDocument(sample_identifier=sample.get("Sample ID")),  # type: ignore[arg-type]
                                device_control_aggregate_document=DeviceControlAggregateDocument(
                                    device_control_document=[
                                        DeviceControlDocumentItem(
                                            device_type="brightfield imager (cell counter)",
                                            detection_type="brightfield",
                                        )
                                    ]
                                ),
                                processed_data_document=ProcessedDataDocument(
                                    data_processing_document=DataProcessingDocument(
                                        cell_type_processing_method=sample.get("Cell type"),  # type: ignore[arg-type]
                                        minimum_cell_diameter=TQuantityValueMicrometer(
                                            value=sample.get("Minimum Diameter (μm)"),  # type: ignore[arg-type]
                                        ),
                                        maximum_cell_diameter=TQuantityValueMicrometer(
                                            value=sample.get("Maximum Diameter (μm)"),  # type: ignore[arg-type]
                                        ),
                                        cell_density_dilution_factor=TQuantityValueUnitless(
                                            value=sample.get("Dilution"),  # type: ignore[arg-type]
                                        ),
                                    ),
                                    viability__cell_counter_=TQuantityValuePercent(
                                        value=sample.get("Viability (%)"),  # type: ignore[arg-type]
                                    ),
                                    viable_cell_density__cell_counter_=TQuantityValueMillionCellsPerMilliliter(
                                        value=sample.get("Viable (x10^6) cells/mL"),  # type: ignore[arg-type]
                                    ),
                                    total_cell_count=TQuantityValueCell(
                                        value=sample.get("Cell count"),  # type: ignore[arg-type]
                                    ),
                                    total_cell_density__cell_counter_=TQuantityValueMillionCellsPerMilliliter(
                                        value=sample.get("Total (x10^6) cells/mL"),  # type: ignore[arg-type]
                                    ),
                                    average_total_cell_diameter=TQuantityValueMicrometer(
                                        value=sample.get("Average diameter (μm)"),  # type: ignore[arg-type]
                                    ),
                                    average_live_cell_diameter__cell_counter_=TQuantityValueMicrometer(
                                        value=sample.get("Average viable diameter (μm)"),  # type: ignore[arg-type]
                                    ),
                                    viable_cell_count=TQuantityValueCell(
                                        value=sample.get("Viable cells"),  # type: ignore[arg-type]
                                    ),
                                    average_total_cell_circularity=TQuantityValueUnitless(
                                        value=sample.get("Average circularity"),  # type: ignore[arg-type]
                                    ),
                                    average_viable_cell_circularity=TQuantityValueUnitless(
                                        value=sample.get("Average viable circularity"),  # type: ignore[arg-type]
                                    ),
                                ),
                            )
                        ],
                    ),
                )
            ],
        ),
    )



from io import BytesIO

from allotropy.parsers.lines_reader import LinesReader


def get_input_lines() -> list[str]:
    return [
        "data section",
        "col1, col2, col3",
        ", ,",
        ", ,",
        ", ,",
        "",
        "header section",
        "element1",
        "element2",
        "element3",
        "",
        "",
        "information",
        "name",
        "123",
        "",
    ]


def get_input_text() -> str:
    return "\n".join(get_input_lines())


def get_input_stream() -> BytesIO:
    return BytesIO(get_input_text().encode("UTF-8"))


def get_test_reader() -> LinesReader:
    return LinesReader(get_input_stream())


def test_reader_constructure() -> None:
    test_reader = get_test_reader()
    assert test_reader.contents == get_input_text()
    assert test_reader.lines == get_input_lines()
    assert test_reader.n_lines == 16
    assert test_reader.current_line == 0


def test_current_line_exists() -> None:
    test_reader = get_test_reader()
    test_reader.current_line = -1
    assert not test_reader.current_line_exists()

    test_reader.current_line = 1
    assert test_reader.current_line_exists()

    test_reader.current_line = 16
    assert not test_reader.current_line_exists()


def test_reader_get() -> None:
    test_reader = get_test_reader()
    test_reader.current_line = 0
    assert test_reader.get() == "data section"

    test_reader.current_line = 6
    assert test_reader.get() == "header section"

    test_reader.current_line = 16
    assert test_reader.get() is None


def test_reader_match() -> None:
    test_reader = get_test_reader()
    test_reader.current_line = 6
    assert test_reader.match("^header section")

    test_reader.current_line = 12
    assert test_reader.match("^information")

    test_reader.current_line = 0
    assert not test_reader.match("^BAD PATTERN$")


def test_reader_is_empty() -> None:
    test_reader = get_test_reader()
    test_reader.current_line = 1
    assert not test_reader.is_empty()

    test_reader.current_line = 5
    assert test_reader.is_empty()


def test_reader_pop() -> None:
    test_reader = get_test_reader()
    test_reader.current_line = 6
    assert test_reader.pop() == "header section"
    assert test_reader.current_line == 7

    test_reader.current_line = 15
    assert test_reader.pop() == ""
    assert test_reader.current_line == 16
    assert test_reader.pop() is None
    assert test_reader.current_line == 16


def test_reader_pop_data() -> None:
    test_reader = get_test_reader()
    test_reader.current_line = 10
    assert test_reader.pop_data() == "information"
    assert test_reader.current_line == 13

    test_reader.current_line = 16
    assert test_reader.pop_data() is None
    assert test_reader.current_line == 16


def test_reader_drop_until() -> None:
    test_reader = get_test_reader()
    assert test_reader.drop_until("^header section") == "header section"
    assert test_reader.current_line == 6

    assert test_reader.drop_until("^BAD PATTERN$") is None
    assert test_reader.current_line == 16


def test_reader_drop_empty() -> None:
    test_reader = get_test_reader()
    test_reader.current_line = 10
    assert test_reader.drop_empty() == "information"
    assert test_reader.current_line == 12

    test_reader.current_line = 15
    assert test_reader.drop_empty() is None
    assert test_reader.current_line == 16


def test_reader_drop_until_empty() -> None:
    test_reader = get_test_reader()
    assert test_reader.drop_until_empty() == ""
    assert test_reader.current_line == 5

    test_reader.current_line = 16
    assert test_reader.drop_until_empty() is None
    assert test_reader.current_line == 16


def test_reader_pop_until() -> None:
    input_lines = get_input_lines()
    test_reader = get_test_reader()
    assert list(test_reader.pop_until("^header section")) == input_lines[:6]


def test_reader_pop_until_empty() -> None:
    input_lines = get_input_lines()
    test_reader = get_test_reader()
    assert list(test_reader.pop_until_empty()) == input_lines[:5]



from dataclasses import dataclass

from allotropy.allotrope.models.shared.definitions.definitions import (
    TNullableQuantityValueWithOptionalUnit,
    TQuantityValueWithOptionalUnit,
)
from allotropy.allotrope.models.shared.definitions.units import (
    Cell,
    Centimeter,
    CubicMillimeter,
    DegreeCelsius,
    GramPerLiter,
    Hertz,
    Microliter,
    Micrometer,
    MilliAbsorbanceUnit,
    Millimeter,
    MillimeterOfMercury,
    MillimolePerLiter,
    MillionCellsPerMilliliter,
    MilliOsmolesPerKilogram,
    MilliSecond,
    Nanometer,
    Number,
    NumberPerMicroliter,
    OpticalDensity,
    Percent,
    PH,
    PicogramPerMilliliter,
    RelativeFluorescenceUnit,
    RelativeLightUnit,
    SecondTime,
    TODO,
    Unitless,
    UnitPerLiter,
)


@dataclass
class TQuantityValueCentimeter(Centimeter, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueMillimeter(Millimeter, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueNanometer(Nanometer, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueMicrometer(Micrometer, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueHertz(Hertz, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueCubicMillimeter(CubicMillimeter, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueUnitless(Unitless, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueSecondTime(SecondTime, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValuePercent(Percent, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueCell(Cell, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueMicroliter(Microliter, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueNumberPerMicroliter(
    NumberPerMicroliter, TQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TQuantityValueMilliSecond(MilliSecond, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueMillionCellsPerMilliliter(
    MillionCellsPerMilliliter, TQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TQuantityValueTODO(TODO, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueGramPerLiter(GramPerLiter, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueDegreeCelsius(DegreeCelsius, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueNumber(Number, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValuePicogramPerMilliliter(
    PicogramPerMilliliter, TQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TQuantityValueUnitPerLiter(UnitPerLiter, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueMillimeterOfMercury(
    MillimeterOfMercury, TQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TQuantityValueOpticalDensity(OpticalDensity, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValuePH(PH, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TQuantityValueMilliOsmolesPerKilogram(
    MilliOsmolesPerKilogram, TQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TQuantityValueMillimolePerLiter(
    MillimolePerLiter, TQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TQuantityValueMilliAbsorbanceUnit(
    MilliAbsorbanceUnit, TQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TRelativeFluorescenceUnit(
    RelativeFluorescenceUnit, TQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TRelativeLightUnit(RelativeLightUnit, TQuantityValueWithOptionalUnit):
    pass


@dataclass
class TNullableQuantityValueCell(Cell, TNullableQuantityValueWithOptionalUnit):
    pass


@dataclass
class TNullableQuantityValueGramPerLiter(
    GramPerLiter, TNullableQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TNullableQuantityValueMicrometer(
    Micrometer, TNullableQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TNullableQuantityValueMilliOsmolesPerKilogram(
    MilliOsmolesPerKilogram, TNullableQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TNullableQuantityValueMillimeterOfMercury(
    MillimeterOfMercury, TNullableQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TNullableQuantityValueMillimolePerLiter(
    MillimolePerLiter, TNullableQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TNullableQuantityValueMillionCellsPerMilliliter(
    MillionCellsPerMilliliter, TNullableQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TNullableQuantityValueOpticalDensity(
    OpticalDensity, TNullableQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TNullableQuantityValuePH(PH, TNullableQuantityValueWithOptionalUnit):
    pass


@dataclass
class TNullableQuantityValuePercent(Percent, TNullableQuantityValueWithOptionalUnit):
    pass


@dataclass
class TNullableQuantityValueTODO(TODO, TNullableQuantityValueWithOptionalUnit):
    pass


@dataclass
class TNullableQuantityValueUnitPerLiter(
    UnitPerLiter, TNullableQuantityValueWithOptionalUnit
):
    pass


@dataclass
class TNullableQuantityValueUnitless(Unitless, TNullableQuantityValueWithOptionalUnit):
    pass



from collections.abc import Iterator
from io import IOBase, StringIO
from re import search
from typing import Optional

import chardet
import pandas as pd

from allotropy.allotrope.allotrope import AllotropyError


class LinesReader:
    def __init__(self, io_: IOBase, encoding: Optional[str] = "UTF-8"):
        stream_contents = io_.read()
        self.raw_contents = (
            self._decode(stream_contents, encoding)
            if isinstance(stream_contents, bytes)
            else stream_contents
        )
        self.contents = self.raw_contents.replace("\r\n", "\n")
        self.lines: list[str] = self.contents.split("\n")
        self.n_lines = len(self.lines)
        self.current_line = 0

    def _decode(self, bytes_content: bytes, encoding: Optional[str]) -> str:
        if not encoding:
            encoding = chardet.detect(bytes_content)["encoding"]
            if not encoding:
                error = "Unable to detect input file encoding"
                raise AllotropyError(error)
        return bytes_content.decode(encoding)

    def current_line_exists(self) -> bool:
        return 0 <= self.current_line < self.n_lines

    def get(self) -> Optional[str]:
        return self.lines[self.current_line] if self.current_line_exists() else None

    def match(self, pattern: str) -> bool:
        line = self.get()
        return False if line is None else bool(search(pattern, line))

    def is_empty(self) -> bool:
        return self.match("^\\s*$")

    def pop(self) -> Optional[str]:
        line = self.get()
        if line is not None:
            self.current_line += 1
        return line

    def pop_if_match(self, pattern: str) -> Optional[str]:
        if self.match(pattern):
            return self.pop()
        return None

    def pop_data(self) -> Optional[str]:
        self.drop_empty()
        return self.pop()

    def drop_until(self, pattern: str) -> Optional[str]:
        while self.current_line_exists() and not self.match(pattern):
            self.pop()
        return self.get()

    def drop_empty(self) -> Optional[str]:
        while self.current_line_exists() and self.is_empty():
            self.pop()
        return self.get()

    def drop_until_empty(self) -> Optional[str]:
        while self.current_line_exists() and not self.is_empty():
            self.pop()
        return self.get()

    def pop_until(self, pattern: str) -> Iterator[str]:
        while self.current_line_exists() and not self.match(pattern):
            line = self.pop()
            if line is not None:
                yield line

    def pop_until_empty(self) -> Iterator[str]:
        while self.current_line_exists() and not self.is_empty():
            line = self.pop()
            if line is not None:
                yield line


class CsvReader(LinesReader):
    def pop_csv_block(self, pattern: Optional[str] = None) -> pd.DataFrame:
        self.drop_empty()
        if pattern:
            if not self.match(pattern):
                msg = f"Did not find {pattern}"
                raise AllotropyError(msg)
            self.pop()  # remove title

        csv_stream = StringIO("\n".join(self.pop_until_empty()))
        self.drop_empty()

        return pd.read_csv(csv_stream, header=None)

    def drop_sections(self, pattern: str) -> None:
        self.drop_empty()
        while True:
            if not self.match(pattern):
                return
            self.drop_until_empty()
            self.drop_empty()



import pytest

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parser_factory import Vendor
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

VENDOR_TYPE = Vendor.AGILENT_GEN5


@pytest.mark.parametrize(
    "filename",
    [
        "endpoint_pathlength_correct_singleplate",
        "endpoint_stdcurve_singleplate",
        "endpoint_stdcurve_singleplate_2",
        "endpoint_stdcurve_multiplate",
        "kinetic_helper_gene_growth_curve",
        "kinetic_singleplate",
        "kinetic_multiplate",
    ],
)
def test_to_allotrope_absorbance(filename: str) -> None:
    test_filepath = f"tests/parsers/agilent_gen5/testdata/absorbance/{filename}.txt"
    expected_filepath = (
        f"tests/parsers/agilent_gen5/testdata/absorbance/{filename}.json"
    )
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(
        allotrope_dict,
        "ultraviolet-absorbance/BENCHLING/2023/09/ultraviolet-absorbance.json",
    )
    validate_contents(allotrope_dict, expected_filepath)


@pytest.mark.parametrize(
    "filename",
    [
        "endpoint_singleplate",
        "endpoint_multiplate",
    ],
)
def test_to_allotrope_fluorescence(filename: str) -> None:
    test_filepath = f"tests/parsers/agilent_gen5/testdata/fluorescence/{filename}.txt"
    expected_filepath = (
        f"tests/parsers/agilent_gen5/testdata/fluorescence/{filename}.json"
    )
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, "fluorescence/BENCHLING/2023/09/fluorescence.json")
    validate_contents(allotrope_dict, expected_filepath)


@pytest.mark.parametrize(
    "filename",
    [
        "endpoint_singleplate",
        "endpoint_multiplate",
    ],
)
def test_to_allotrope_luminescence(filename: str) -> None:
    test_filepath = f"tests/parsers/agilent_gen5/testdata/luminescence/{filename}.txt"
    expected_filepath = (
        f"tests/parsers/agilent_gen5/testdata/luminescence/{filename}.json"
    )
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, "luminescence/BENCHLING/2023/09/luminescence.json")
    validate_contents(allotrope_dict, expected_filepath)


def test_to_allotrope_invalid_plate_data() -> None:
    with pytest.raises(AllotropeConversionError, match="No plate data found in file"):
        from_file("tests/parsers/agilent_gen5/testdata/garbage.txt", VENDOR_TYPE)



from typing import Optional

import pandas as pd
import pytest

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parsers.novabio_flex2.constants import (
    MOLAR_CONCENTRATION_CLS_BY_UNIT,
    PROPERTY_MAPPINGS,
)
from allotropy.parsers.novabio_flex2.novabio_flex2_structure import (
    Analyte,
    Data,
    Sample,
    SampleList,
    Title,
)
from tests.parsers.novabio_flex2.novabio_flex2_data import (
    get_data,
    get_input_stream,
    get_input_title,
)


@pytest.mark.parametrize(
    "filename,processing_time,device_identifier",
    [
        ("SampleResults2022-06-28_142558.csv", "2022-06-28 142558", None),
        (
            "SampleResultsT26918070C2021-02-18_104838.csv",
            "2021-02-18 104838",
            "T26918070C",
        ),
    ],
)
@pytest.mark.short
def test_create_title(
    filename: str, processing_time: str, device_identifier: Optional[str]
) -> None:
    title = Title.create(filename)
    assert title.processing_time == processing_time
    assert title.device_identifier == device_identifier


@pytest.mark.parametrize(
    "filename",
    (
        "invalid_filename",  # no extension
        "T26918070C2021-02-18_104838.csv",  # filename does not start with SampleResults
        "SampleResults2021-02-18_104838T26918070C.csv",  # wrong order of timestamp and identifier
    ),
)
@pytest.mark.short
def test_create_title_invalid_filename(filename: str) -> None:
    with pytest.raises(AllotropeConversionError):
        Title.create(filename)


@pytest.mark.short
def test_create_analyte() -> None:
    nh4_analyte = Analyte.create("NH4+", 100)
    assert nh4_analyte.name == "ammonium"
    assert nh4_analyte.molar_concentration == MOLAR_CONCENTRATION_CLS_BY_UNIT["mmol/L"](
        value=100
    )

    gluc_analyte = Analyte.create("Gluc", 1.1)
    assert gluc_analyte.name == "glucose"
    assert gluc_analyte.molar_concentration == MOLAR_CONCENTRATION_CLS_BY_UNIT["g/L"](
        value=1.1
    )


@pytest.mark.short
def test_create_invalid_analyte() -> None:
    with pytest.raises(AllotropeConversionError):
        Analyte.create("FAKE", 100)


@pytest.mark.short
def test_create_sample() -> None:
    data = {
        "Sample ID": "BP_R10_KP_008_D0",
        "Sample Type": "Spent Media",
        "Date & Time": pd.Timestamp("2022-06-24 14:34:52"),
        "Batch ID": "KP_008",
        "Operator": "Kermit",
        "Gln": 1.83,
        "Ca++": 0.82,
        "O2 Saturation": 100.0,
        "CO2 Saturation": 0,  # zero here makes sure we allow falsey values
        "Unmapped column": 5,
    }
    sample = Sample.create(pd.Series(data=data))

    assert sample.identifier == "BP_R10_KP_008_D0"
    assert sample.role_type == "Spent Media"
    assert sample.measurement_time == "2022-06-24T14:34:52"
    assert sample.batch_identifier == "KP_008"
    assert sorted(sample.analytes) == sorted(
        [
            Analyte.create("Gln", 1.83),
            Analyte.create("Ca++", 0.82),
        ]
    )
    assert sample.properties == {
        "co2_saturation": PROPERTY_MAPPINGS["co2_saturation"]["cls"](value=0),
        "o2_saturation": PROPERTY_MAPPINGS["o2_saturation"]["cls"](value=100.0),
    }


@pytest.mark.short
def test_sample_sorting() -> None:
    analytes = [
        Analyte.create("Gln", 1.83),
        Analyte.create("Glu", 0.33),
        Analyte.create("Gluc", 2.65),
        Analyte.create("Lac", 0.18),
        Analyte.create("NH4+", 0.48),
    ]
    assert sorted(analytes) == [
        Analyte.create("NH4+", 0.48),
        Analyte.create("Gluc", 2.65),
        Analyte.create("Glu", 0.33),
        Analyte.create("Gln", 1.83),
        Analyte.create("Lac", 0.18),
    ]


@pytest.mark.short
def test_create_sample_list() -> None:
    sample_list = SampleList.create(
        pd.DataFrame(
            {
                "Sample ID": ["SAMPLE_1", "SAMPLE_2"],
                "Sample Type": ["Spent Media", "Spent Media"],
                "Date & Time": [
                    pd.Timestamp("2022-06-24 14:34:52"),
                    pd.Timestamp("2022-06-24 14:34:52"),
                ],
                "Operator": ["Kermit", "Other"],
            }
        )
    )

    assert sample_list.analyst == "Kermit"
    assert len(sample_list.samples) == 2
    assert sample_list.samples[0].identifier == "SAMPLE_1"
    assert sample_list.samples[1].identifier == "SAMPLE_2"


@pytest.mark.short
def test_create_sample_list_invalid_no_samples() -> None:
    with pytest.raises(AllotropeConversionError):
        SampleList.create(pd.DataFrame({}))


@pytest.mark.short
def test_create_sample_list_invalid_no_analyst() -> None:
    with pytest.raises(AllotropeConversionError):
        SampleList.create(
            pd.DataFrame(
                {
                    "Sample ID": ["SAMPLE_1", "SAMPLE_2"],
                    "Sample Type": ["Spent Media", "Spent Media"],
                    "Date & Time": [
                        pd.Timestamp("2022-06-24 14:34:52"),
                        pd.Timestamp("2022-06-24 14:34:52"),
                    ],
                }
            )
        )


@pytest.mark.short
def test_create_data() -> None:
    assert Data.create(get_input_stream(), get_input_title()) == get_data()



import pytest

from allotropy.parser_factory import Vendor
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_parser import (
    PerkinElmerEnvisionParser,
)
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from tests.parsers.perkin_elmer_envision.perkin_elmer_envision_data import (
    get_data,
    get_model,
)
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

output_files = (
    "PE_Envision_fluorescence_example01",
    "PE_Envision_fluorescence_example02",
    "PE_Envision_fluorescence_example03",
)

VENDOR_TYPE = Vendor.PERKIN_ELMER_ENVISION


@pytest.mark.parametrize("output_file", output_files)
def test_parse_perkin_elmer_envision_to_asm(output_file: str) -> None:
    test_filepath = f"tests/parsers/perkin_elmer_envision/testdata/{output_file}.csv"
    expected_filepath = (
        f"tests/parsers/perkin_elmer_envision/testdata/{output_file}.json"
    )
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, "fluorescence/BENCHLING/2023/09/fluorescence.json")
    validate_contents(allotrope_dict, expected_filepath)


@pytest.mark.short
def test_get_model() -> None:
    parser = PerkinElmerEnvisionParser(TimestampParser())
    model = parser._get_model(get_data())

    if model.measurement_aggregate_document:
        model.measurement_aggregate_document.measurement_identifier = ""

    assert model == get_model()



from typing import Any, ClassVar, Optional
import uuid

from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueNanometer,
    TQuantityValueNumber,
)
from allotropy.parsers.moldev_softmax_pro.plate_block import (
    PlateBlock,
    split_wavelengths,
    WellData,
)
from allotropy.parsers.utils.values import assert_not_none, natural_sort_key, try_int


class FluorescenceOrLuminescencePlateBlock(PlateBlock):
    EXCITATION_WAVELENGTHS_IDX: ClassVar[int]

    def parse_read_mode_header(self, header: list[Optional[str]]) -> None:
        [
            excitation_wavelengths_str,
            self.cutoff,
            cutoff_filters_str,
            self.sweep_wave,
            self.sweep_wavelength,
            self.reads_per_well,
            self.pmt_gain,
            self.start_integration_time,
            self.end_integration_time,
            self.first_row,
            num_rows,
        ] = header[
            self.EXCITATION_WAVELENGTHS_IDX : self.EXCITATION_WAVELENGTHS_IDX + 11
        ]
        self.excitation_wavelengths = split_wavelengths(excitation_wavelengths_str)
        self.cutoff_filters = split_wavelengths(cutoff_filters_str)
        self.num_rows = assert_not_none(try_int(num_rows), "num_rows")


class FluorescencePlateBlock(FluorescenceOrLuminescencePlateBlock):
    READ_MODE = "Fluorescence"
    CONCEPT = "fluorescence"
    UNIT = "RFU"
    DATA_TYPE_IDX = 7
    EXCITATION_WAVELENGTHS_IDX = 20

    # TODO: the reason we can't factor out DeviceControlDocumentItem and the enclosing classes is because the
    # Fluorescence ASM model has these extra fields. We may be able to fix this by templating the PlateBlock
    # class on these classes, or some clever shared/models refactoring.
    def generate_device_control_doc(self) -> DeviceControlDocumentItem:
        device_control_doc = DeviceControlDocumentItem(
            detector_gain_setting=self.pmt_gain
        )

        if self.is_single_wavelength:
            device_control_doc.detector_wavelength_setting = TQuantityValueNanometer(
                self.wavelengths[0]
            )
        if self.excitation_wavelengths and len(self.excitation_wavelengths) == 1:
            device_control_doc.excitation_wavelength_setting = TQuantityValueNanometer(
                self.excitation_wavelengths[0]
            )
        if self.cutoff_filters and len(self.cutoff_filters) == 1:
            device_control_doc.wavelength_filter_cutoff_setting = (
                TQuantityValueNanometer(self.cutoff_filters[0])
            )

        return device_control_doc

    def generate_measurement_doc(
        self, well: str, well_data: WellData
    ) -> MeasurementDocumentItem:
        measurement = MeasurementDocumentItem(
            DeviceControlAggregateDocument([self.generate_device_control_doc()]),
            self.generate_sample_document(well),
        )

        if well_data.temperature is not None:
            measurement.compartment_temperature = TQuantityValueDegreeCelsius(
                float(well_data.temperature)
            )

        if not well_data.is_empty:
            measurement.data_cube = self.generate_data_cube(well_data)

        if well_data.processed_data:
            measurement.processed_data_aggregate_document = (
                self.generate_processed_data_aggreate_document(well_data)
            )

        return measurement

    def to_allotrope(self) -> Any:
        wells = sorted(self.well_data.keys(), key=natural_sort_key)

        allotrope_file = Model(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_identifier=str(uuid.uuid4()),
                plate_well_count=TQuantityValueNumber(self.num_wells),
                measurement_document=[
                    self.generate_measurement_doc(well, self.well_data[well])
                    for well in wells
                ],
            )
        )

        return allotrope_file



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Optional

import pandas as pd

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parsers.roche_cedex_bioht.constants import (
    MOLAR_CONCENTRATION_CLS_BY_UNIT,
    NON_AGGREGABLE_PROPERTIES,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_reader import (
    RocheCedexBiohtReader,
)


@dataclass
class Title:
    data_processing_time: Optional[str]
    analyst: str
    model_number: Optional[str]
    device_serial_number: Optional[str]

    @staticmethod
    def create(title_data: pd.Series) -> Title:
        analyst = title_data.get("analyst")
        if analyst is None:
            msg = "Unable to obtain analyst from input data"
            raise AllotropeConversionError(msg)

        device_serial_number = title_data.get("device serial number")
        if device_serial_number is None:
            msg = "Unable to obtain device serial number from input data"
            raise AllotropeConversionError(msg)

        return Title(
            title_data.get("data processing time"),  # type: ignore[arg-type]
            analyst,  # type: ignore[arg-type]
            title_data.get("model number"),  # type: ignore[arg-type]
            str(device_serial_number),
        )


@dataclass
class Analyte:
    name: str
    concentration_value: Optional[float]
    unit: Optional[str]

    @staticmethod
    def create(data: pd.Series) -> Analyte:
        analyte_name: str = data.get("analyte name")  # type: ignore[assignment]
        concentration_value: Optional[float] = data.get("concentration value")  # type: ignore[assignment]
        unit: Optional[str] = data.get("concentration unit")  # type: ignore[assignment]

        return Analyte(analyte_name, concentration_value, unit)


@dataclass
class AnalyteList:
    analytes: list[Analyte]
    molar_concentration_dict: dict
    molar_concentration_nans: dict
    non_aggregrable_dict: dict
    non_aggregable_nans: dict
    num_measurement_docs: int

    @staticmethod
    def create(data: pd.DataFrame) -> AnalyteList:
        analytes = [Analyte.create(analyte_data) for _, analyte_data in data.iterrows()]
        molar_concentration_dict = defaultdict(list)
        molar_concentration_nans = {}
        non_aggregrable_dict = defaultdict(list)
        non_aggregable_nans = {}
        num_measurement_docs = 1

        for analyte in analytes:
            analyte_name = analyte.name
            concentration_value = analyte.concentration_value

            if analyte_cls := NON_AGGREGABLE_PROPERTIES.get(analyte_name):
                if concentration_value is None:
                    non_aggregable_nans[analyte_name] = analyte_cls(
                        value=concentration_value
                    )
                else:
                    non_aggregrable_dict[analyte_name].append(
                        analyte_cls(value=concentration_value)
                    )

                num_measurement_docs = max(
                    len(non_aggregrable_dict[analyte_name]), num_measurement_docs
                )
            else:
                unit = analyte.unit
                if unit is None:
                    continue

                molar_concentration_item_cls = MOLAR_CONCENTRATION_CLS_BY_UNIT.get(unit)
                if molar_concentration_item_cls is None:
                    continue

                molar_concentration_item = molar_concentration_item_cls(
                    value=concentration_value
                )
                if concentration_value is None:
                    molar_concentration_nans[analyte_name] = molar_concentration_item
                else:
                    molar_concentration_dict[analyte_name].append(
                        molar_concentration_item
                    )
                num_measurement_docs = max(
                    len(molar_concentration_dict[analyte_name]), num_measurement_docs
                )

        # Only include None values if there is not a valid value for that analyte
        for analyte_name in non_aggregable_nans:
            if len(non_aggregrable_dict[analyte_name]) == 0:
                non_aggregrable_dict[analyte_name].append(
                    non_aggregable_nans[analyte_name]
                )
                num_measurement_docs = max(
                    len(non_aggregrable_dict[analyte_name]), num_measurement_docs
                )

        for analyte_name in molar_concentration_nans:
            if len(molar_concentration_dict[analyte_name]) == 0:
                molar_concentration_dict[analyte_name].append(
                    molar_concentration_nans[analyte_name]
                )
                num_measurement_docs = max(
                    len(molar_concentration_dict[analyte_name]), num_measurement_docs
                )

        return AnalyteList(
            analytes,
            molar_concentration_dict,
            molar_concentration_nans,
            non_aggregrable_dict,
            non_aggregable_nans,
            num_measurement_docs,
        )


@dataclass
class Sample:
    name: str
    role_type: str
    measurement_time: str
    analyte_list: AnalyteList
    batch: Optional[str] = None

    @staticmethod
    def create(name: str, batch: Optional[str], samples_data: pd.DataFrame) -> Sample:
        condition = samples_data["sample identifier"] == name
        condition &= samples_data["batch identifier"] == batch
        sample_data = samples_data[condition].sort_values(by="analyte name")

        role_type = sample_data.iloc[0]["sample role type"]
        measurement_time = str(sample_data.iloc[0]["measurement time"])

        return Sample(
            name,
            role_type,
            measurement_time,
            AnalyteList.create(sample_data),
            batch=batch or None,
        )


@dataclass
class Data:
    title: Title
    samples: list[Sample]

    @staticmethod
    def create(reader: RocheCedexBiohtReader) -> Data:
        # A sample group is defined by both the sample and the batch identifier
        sample_groups = reader.samples_data.groupby(
            ["sample identifier", "batch identifier"]
        ).groups.keys()

        return Data(
            title=Title.create(reader.title_data),
            samples=[Sample.create(name, batch, reader.samples_data) for name, batch in sample_groups],  # type: ignore[has-type, misc]
        )



from allotropy.allotrope.models.luminescence_benchling_2023_09_luminescence import (
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    MeasurementDocumentItem,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValuePicogramPerMilliliter,
)
from allotropy.parsers.agilent_gen5.constants import ReadMode
from allotropy.parsers.agilent_gen5.data_point import DataPoint


class LuminescenceDataPoint(DataPoint):
    read_mode = ReadMode.LUMINESCENCE
    unit = "RLU"

    def to_measurement_doc(self) -> MeasurementDocumentItem:
        measurement_doc = MeasurementDocumentItem(
            DeviceControlAggregateDocument([DeviceControlDocumentItem()]),
            self.generate_sample_doc(),
        )
        measurement_doc.data_cube = self.generate_data_cube()

        if self.concentration:
            measurement_doc.mass_concentration = TQuantityValuePicogramPerMilliliter(
                self.concentration
            )
        if self.processed_data:
            measurement_doc.processed_data_aggregate_document = (
                self.processed_data_doc()
            )
        if self.temperature:
            measurement_doc.compartment_temperature = TQuantityValueDegreeCelsius(
                float(self.temperature)
            )
        return measurement_doc



from allotropy.allotrope.schema_parser.model_class_editor import (
    get_manifest_from_schema_path,
    ModelClassEditor,
)


def test_get_manifest_from_schema_path() -> None:
    schema_path = "src/allotropy/allotrope/schemas/cell_counting/BENCHLING/2023/09/cell-counting.json"
    expected = "http://purl.allotrope.org/manifests/cell_counting/BENCHLING/2023/09/cell-counting.manifest"
    assert get_manifest_from_schema_path(schema_path) == expected


def test_modify_file() -> None:
    classes_to_skip = {"ClassA"}
    imports_to_add = {"definitions": {"Thing1", "Thing2"}}
    editor = ModelClassEditor("fake_manifest", classes_to_skip, imports_to_add)

    model_file_contents = """
import json
from typing import Union


@dataclass
class ClassA:
    key: str
    value: str


@dataclass
class ClassB:
    key: str
    value: str


@dataclass
class UnusedClass:
    value: int


@dataclass
class Model:
    key: str
    value: str
    b_thing: ClassB
"""

    expected = """
import json
from typing import Union

from allotropy.allotrope.models.shared.definitions import Thing1, Thing2


@dataclass
class ClassB:
    key: str
    value: str


@dataclass
class Model:
    key: str
    value: str
    b_thing: ClassB
    manifest: str = \"fake_manifest\"
"""
    assert editor.modify_file(model_file_contents) == expected



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValuePercent,
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TQuantityValue,
    TStringValue,
)


@dataclass
class PeakItem:
    retention_time: TQuantityValueSecondTime
    peak_end: Optional[TQuantityValueSecondTime] = None
    identifier: Optional[TStringValue] = None
    relative_peak_height: Optional[TQuantityValuePercent] = None
    written_name: Optional[TStringValue] = None
    peak_height: Optional[TQuantityValue] = None
    capacity_factor__chromatography_: Optional[TQuantityValueUnitless] = None
    peak_area: Optional[TQuantityValue] = None
    relative_peak_area: Optional[TQuantityValuePercent] = None
    peak_start: Optional[TQuantityValueSecondTime] = None
    peak_selectivity__chromatography_: Optional[TQuantityValueUnitless] = None

    peak_width_at_4_4___of_height: Optional[TQuantityValueSecondTime] = None
    peak_width_at_13_4___of_height: Optional[TQuantityValueSecondTime] = None
    peak_width_at_32_4___of_height: Optional[TQuantityValueSecondTime] = None
    peak_width_at_60_7___of_height: Optional[TQuantityValueSecondTime] = None
    peak_width_at_half_height: Optional[TQuantityValueSecondTime] = None
    peak_width_at_5___of_height: Optional[TQuantityValueSecondTime] = None
    peak_width_at_baseline: Optional[TQuantityValueSecondTime] = None
    peak_width_at_inflection: Optional[TQuantityValueSecondTime] = None
    peak_width_at_10___of_height: Optional[TQuantityValueSecondTime] = None
    peak_width: Optional[TQuantityValueSecondTime] = None
    statistical_skew__chromatography_: Optional[TQuantityValueUnitless] = None
    asymmetry_factor_measured_at_5___height: Optional[TQuantityValueUnitless] = None
    asymmetry_factor_measured_at_10___height: Optional[TQuantityValueUnitless] = None
    asymmetry_factor_squared_measured_at_10___height: Optional[
        TQuantityValueUnitless
    ] = None
    asymmetry_factor_squared_measured_at_4_4___height: Optional[
        TQuantityValueUnitless
    ] = None
    asymmetry_factor_measured_at_4_4___height: Optional[TQuantityValueUnitless] = None
    chromatographic_peak_asymmetry_factor: Optional[TQuantityValueUnitless] = None
    asymmetry_factor_measured_at_baseline: Optional[TQuantityValueUnitless] = None
    chromatographic_peak_resolution: Optional[TQuantityValueUnitless] = None
    chromatographic_peak_resolution_using_baseline_peak_widths: Optional[
        TQuantityValueUnitless
    ] = None
    chromatographic_peak_resolution_using_peak_width_at_half_height: Optional[
        TQuantityValueUnitless
    ] = None
    chromatographic_peak_resolution_using_statistical_moments: Optional[
        TQuantityValueUnitless
    ] = None
    number_of_theoretical_plates__chromatography_: Optional[
        TQuantityValueUnitless
    ] = None
    number_of_theoretical_plates_measured_at_60_7___of_peak_height: Optional[
        TQuantityValueUnitless
    ] = None
    number_of_theoretical_plates_measured_at_32_4___of_peak_height: Optional[
        TQuantityValueUnitless
    ] = None
    number_of_theoretical_plates_measured_at_13_4___of_peak_height: Optional[
        TQuantityValueUnitless
    ] = None
    number_of_theoretical_plates_measured_at_4_4___of_peak_height: Optional[
        TQuantityValueUnitless
    ] = None
    number_of_theoretical_plates_by_tangent_method: Optional[
        TQuantityValueUnitless
    ] = None
    number_of_theoretical_plates_by_peak_width_at_half_height: Optional[
        TQuantityValueUnitless
    ] = None
    number_of_theoretical_plates_by_peak_width_at_half_height__JP14_: Optional[
        TQuantityValueUnitless
    ] = None

    def __post_init__(self) -> None:
        any_of_keys = [
            "peak_width_at_4_4___of_height",
            "peak_width_at_13_4___of_height",
            "peak_width_at_32_4___of_height",
            "peak_width_at_60_7___of_height",
            "peak_width_at_half_height",
            "peak_width_at_5___of_height",
            "peak_width_at_baseline",
            "peak_width_at_inflection",
            "peak_width_at_10___of_height",
            "peak_width",
            "statistical_skew__chromatography_",
            "asymmetry_factor_measured_at_5___height",
            "asymmetry_factor_measured_at_10___height",
            "asymmetry_factor_squared_measured_at_10___height",
            "asymmetry_factor_squared_measured_at_4_4___height",
            "asymmetry_factor_measured_at_4_4___height",
            "chromatographic_peak_asymmetry_factor",
            "asymmetry_factor_measured_at_baseline",
            "chromatographic_peak_resolution",
            "chromatographic_peak_resolution_using_baseline_peak_widths",
            "chromatographic_peak_resolution_using_peak_width_at_half_height",
            "chromatographic_peak_resolution_using_statistical_moments",
            "number_of_theoretical_plates__chromatography_",
            "number_of_theoretical_plates_measured_at_60_7___of_peak_height",
            "number_of_theoretical_plates_measured_at_32_4___of_peak_height",
            "number_of_theoretical_plates_measured_at_13_4___of_peak_height",
            "number_of_theoretical_plates_measured_at_4_4___of_peak_height",
            "number_of_theoretical_plates_by_tangent_method",
            "number_of_theoretical_plates_by_peak_width_at_half_height",
            "number_of_theoretical_plates_by_peak_width_at_half_height__JP14_",
        ]
        # Logic for enforcing anyOf
        if all(getattr(self, key) is None for key in any_of_keys):
            error = f"At least one of {any_of_keys} must be set."
            raise ValueError(error)


@dataclass
class PeakList:
    peak: Optional[list[PeakItem]] = None



from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import WellItem
from allotropy.parsers.appbio_quantstudio.views import View


class SampleView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        items: dict[str, list[WellItem]] = {}
        for well_item in well_items:
            if well_item.sample_identifier not in items:
                items[well_item.sample_identifier] = []
            items[well_item.sample_identifier].append(well_item)
        return items


class TargetView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        items: dict[str, list[WellItem]] = {}
        for well_item in well_items:
            if well_item.target_dna_description not in items:
                items[well_item.target_dna_description] = []
            items[well_item.target_dna_description].append(well_item)
        return items


class TargetRoleView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        items: dict[str, list[WellItem]] = {}
        for well_item in well_items:
            if well_item.sample_role_type == "STANDARD":
                if well_item.target_dna_description not in items:
                    items[well_item.target_dna_description] = []
                items[well_item.target_dna_description].append(well_item)
        return items



from collections import defaultdict
from dataclasses import dataclass
import io
import json
import os
from pathlib import Path
import re
from typing import Optional

from allotropy.allotrope.schema_parser.schema_model import (
    get_all_schema_components,
    get_schema_definitions_mapping,
)

SCHEMA_DIR_PATH = "src/allotropy/allotrope/schemas"
SHARED_FOLDER_MODULE = "allotropy.allotrope.models.shared"


def get_shared_schema_info(schema_path: str) -> tuple[set[str], dict[str, set[str]]]:
    with open(schema_path) as f:
        schema = json.load(f)

    classes_to_skip = set()
    imports_to_add = defaultdict(set)

    # Get all properties from the schema and all definitions.
    schema_mapping = get_all_schema_components(schema)
    if "$defs" in schema:
        schema_mapping.update(schema["$defs"])
    if "$custom" in schema:
        schema_mapping.update(schema["$custom"])

    # Get schemas defined in models/shared
    shared_schema_mapping = get_schema_definitions_mapping()

    # Match schemas to get classes to remove from the generated code and imports to add.
    for name, component_schema in schema_mapping.items():
        for schema_model in shared_schema_mapping.get(name, []):
            # TODO: log warning visible in script output if schema name is found but no schemas match.
            if schema_model.schema == component_schema:
                classes_to_skip.add(schema_model.import_info[1])
                imports_to_add[schema_model.import_info[0]].add(
                    schema_model.import_info[1]
                )

    return classes_to_skip, dict(imports_to_add)


def get_manifest_from_schema_path(schema_path: str) -> str:
    relpath = Path(os.path.relpath(schema_path, SCHEMA_DIR_PATH))
    return (
        f"http://purl.allotrope.org/manifests/{relpath.parent}/{relpath.stem}.manifest"
    )


@dataclass
class ClassLines:
    """Represents a set of lines defining a class."""

    lines: list[str]

    @property
    def class_name(self) -> str:
        match = None
        if self.lines[0].startswith("@dataclass"):
            match = re.match("class ([^\\(:]*)", self.lines[1])
        elif self.lines[0].startswith("class"):
            match = re.match("class ([^\\(:]*)", self.lines[0])
        elif " = " in self.lines[0]:
            match = re.match("(\\S+) =", self.lines[0])
        if not match:
            error = f"Could not determine class name for : {''.join(self.lines)}"
            raise ValueError(error)
        return match.groups()[0]


class ModelClassEditor:
    """
    Iterates over a file of generated models and rewrites it with modifications, including:

    - Adding manifest to base model
    - Factoring out models found in shared/models
    """

    def __init__(
        self,
        manifest: str,
        classes_to_skip: set[str],
        imports_to_add: dict[str, set[str]],
    ):
        self.manifest = manifest
        self.classes_to_skip = classes_to_skip
        self.imports_to_add = imports_to_add

    def _handle_class_lines(self, class_lines: ClassLines) -> Optional[list[str]]:
        # Given a ClassLines object, returns the lines to add to the file.
        if class_lines.class_name in self.classes_to_skip:
            return None
        # Add manifest to base Model class.
        if class_lines.class_name == "Model" and "manifest" not in "".join(
            class_lines.lines
        ):
            class_lines.lines.append(f'    manifest: str = "{self.manifest}"\n')
        return class_lines.lines

    def _get_class_lines(
        self, file: io.TextIOBase, existing_lines: Optional[list[str]] = None
    ) -> Optional[ClassLines]:
        # Reads lines for the next class and returns as a ClassLines object.
        lines: list[str] = existing_lines or []
        started = False
        while True:
            line = file.readline()
            # Skip empty lines at the start, return when we hit a newline after the class.
            if line == "\n":
                if started:
                    return ClassLines(lines)
                else:
                    continue
            # End of file, return the class if we've read one in on this run.
            if not line:
                return ClassLines(lines) if started else None
            started = True
            lines.append(line)

    def modify_file(self, file_contents: str) -> str:
        new_contents: list[str] = []

        f = io.StringIO(file_contents)

        # Scan past comments and import lines.
        while True:
            line = f.readline()
            # TODO: this needs work to be more robust.
            if (
                line == "\n"
                or line.startswith("#")
                or line.startswith("from")
                or line.startswith("import")
            ):
                new_contents.append(line)
            else:
                break

        # We read the first line in the first class, so we need to add it back in
        existing_lines = [line]

        # Remove one newline after default imports
        if self.imports_to_add:
            new_contents = new_contents[:-1]
            # Add new imports
            for module, classes in self.imports_to_add.items():
                new_contents.append(
                    f"from {SHARED_FOLDER_MODULE}.{module} import {', '.join(sorted(classes))}\n"
                )
        else:
            new_contents = new_contents[:-2]

        # Parse classes
        all_classes: set[str] = set()
        while True:
            class_lines = self._get_class_lines(f, existing_lines)
            existing_lines = []
            if not class_lines:
                break
            updated_lines = self._handle_class_lines(class_lines)
            if updated_lines:
                new_contents.extend(["\n", "\n", *updated_lines])
                all_classes.add(class_lines.class_name)

        # If there are any unused classes, remove them.
        new = "".join(new_contents)
        classes_to_remove: set[str] = set()
        for class_name in all_classes:
            if class_name == "Model":
                continue
            if not re.search(rf"(?<!class )(?<!\w){class_name}(?!\w)(?! = )", new):
                classes_to_remove.add(class_name)

        if classes_to_remove:
            self.classes_to_skip.update(classes_to_remove)
            self.imports_to_add = {}
            return self.modify_file(new)

        return new


def modify_file(model_path: str, schema_path: str) -> None:
    classes_to_skip, imports_to_add = get_shared_schema_info(schema_path)
    manifest = get_manifest_from_schema_path(schema_path)

    editor = ModelClassEditor(manifest, classes_to_skip, imports_to_add)

    with open(model_path) as f:
        contents = f.read()

    with open(model_path, "w") as f:
        f.write(editor.modify_file(contents))



# generated by datamodel-codegen:
#   filename:  liquid-chromatography.json
#   timestamp: 2023-09-18T15:35:45+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

from allotropy.allotrope.models.shared.components.liquid_chromatography import (
    PeakList,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCentimeter,
    TQuantityValueCubicMillimeter,
    TQuantityValueHertz,
    TQuantityValueMicrometer,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDatacube,
    TDateTimeValue,
    TQuantityValue,
    TStringValue,
)


@dataclass
class DeviceDocumentItem:
    device_type: TStringValue
    device_identifier: Optional[TStringValue] = None
    model_number: Optional[TStringValue] = None
    field_index: Optional[int] = None


@dataclass
class DeviceSystemDocument:
    asset_management_identifier: TStringValue
    description: Optional[Any] = None
    brand_name: Optional[TStringValue] = None
    product_manufacturer: Optional[TStringValue] = None
    pump_model_number: Optional[TStringValue] = None
    detector_model_number: Optional[TStringValue] = None
    device_document: Optional[list[DeviceDocumentItem]] = None


@dataclass
class SampleDocument:
    sample_identifier: TStringValue
    description: Optional[Any] = None
    written_name: Optional[TStringValue] = None


@dataclass
class ChromatographyColumnDocument:
    chromatography_column_particle_size: Optional[TQuantityValueMicrometer] = None
    chromatography_column_chemistry_type: Optional[TStringValue] = None
    chromatography_column_serial_number: Optional[TStringValue] = None
    column_inner_diameter: Optional[TQuantityValueMillimeter] = None
    product_manufacturer: Optional[TStringValue] = None
    chromatography_column_length: Optional[TQuantityValueCentimeter] = None
    chromatography_column_part_number: Optional[TStringValue] = None


@dataclass
class DeviceControlDocumentItem:
    device_type: TStringValue
    device_identifier: Optional[TStringValue] = None
    product_manufacturer: Optional[TStringValue] = None
    brand_name: Optional[TStringValue] = None
    equipment_serial_number: Optional[TStringValue] = None
    model_number: Optional[TStringValue] = None
    firmware_version: Optional[TStringValue] = None
    detection_type: Optional[TStringValue] = None
    electronic_absorbance_wavelength_setting: Optional[TQuantityValueNanometer] = None
    electronic_absorbance_bandwidth_setting: Optional[TQuantityValueNanometer] = None
    electronic_absorbance_reference_bandwidth_setting: Optional[
        TQuantityValueNanometer
    ] = None
    electronic_absorbance_reference_wavelength_setting: Optional[
        TQuantityValueNanometer
    ] = None
    detector_offset_setting: Optional[TQuantityValue] = None
    detector_sampling_rate_setting: Optional[TQuantityValueHertz] = None
    field_index: Optional[int] = None


@dataclass
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass
class InjectionDocument:
    autosampler_injection_volume_setting__chromatography_: TQuantityValueCubicMillimeter
    injection_identifier: TStringValue
    injection_time: TDateTimeValue


@dataclass
class ProcessedDataDocument:
    peak_list: Optional[PeakList] = None


@dataclass
class DiagnosticTraceDocumentItem:
    description: Any
    data_cube: Optional[TDatacube] = None


@dataclass
class DiagnosticTraceAggregateDocument:
    diagnostic_trace_document: Optional[list[DiagnosticTraceDocumentItem]] = None


@dataclass
class MeasurementDocumentItem:
    chromatography_column_document: ChromatographyColumnDocument
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    injection_document: InjectionDocument
    detection_type: TStringValue
    chromatogram_data_cube: TDatacube
    measurement_identifier: Optional[TStringValue] = None
    three_dimensional_ultraviolet_spectrum_data_cube: Optional[TDatacube] = None
    processed_data_document: Optional[ProcessedDataDocument] = None
    diagnostic_trace_aggregate_document: Optional[
        DiagnosticTraceAggregateDocument
    ] = None


@dataclass
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocumentItem]


@dataclass
class LiquidChromatographyDocumentItem:
    analyst: TStringValue
    measurement_aggregate_document: MeasurementAggregateDocument
    submitter: Optional[TStringValue] = None


@dataclass
class LiquidChromatographyAggregateDocument:
    device_system_document: DeviceSystemDocument
    liquid_chromatography_document: list[LiquidChromatographyDocumentItem]


@dataclass
class Model:
    liquid_chromatography_aggregate_document: Optional[
        LiquidChromatographyAggregateDocument
    ] = None
    manifest: str = "http://purl.allotrope.org/manifests/liquid-chromatography/REC/2023/03/liquid-chromatography.manifest"
